#pragma once

// A wrapper around the IDirect3DTexture9*.
class CTexture
{
private:
    // The texture itself
    CComPtr<IDirect3DTexture9> m_tex;

    // Texture mipmap filter, is used for both D3DSAMP_MIPFILTER, D3DSAMP_MINFILTER and D3DSAMP_MAGFILTER
    D3DTEXTUREFILTERTYPE m_texMipFilter;
public:
    inline CTexture()
    {
		m_texMipFilter = D3DTEXF_LINEAR;	//< For point sprites, it looks same as D3DTEXF_ANISOTROPIC
    }
    inline ~CTexture(){}

    // Load application-defined resource, <s>containing a whole bmp/dds/dib/hdr/jpg/pfm/png/ppm/tga file</s>
	// Containing a DXT5 DDS texture with mipmaps,
    // into the new D3DPOOL_MANAGED texture.
    HRESULT LoadResource( IDirect3DDevice9* pDevice,
        LPCTSTR lpName, LPCTSTR lpType = NULL,
        HMODULE hModule = NULL )
    {
        CHECK( NULL != pDevice, return E_POINTER, "IDirect3DDevice9* pointer can't be NULL" );
        if( NULL == lpType ) lpType = L"TEXTURES";

        HRSRC rsRC = FindResource( hModule, lpName, lpType );
        CHECK( NULL != rsRC, return GetLastHr(), "Unable to find the resource.\n" );

        const DWORD cbResourceSize = SizeofResource( hModule, rsRC );
        if( 0 == cbResourceSize )
            return GetLastHr();

        HGLOBAL hgRC = ::LoadResource( hModule, rsRC );
        if( NULL == hgRC )
            return GetLastHr();

        const void* pcvRC = LockResource( hgRC );
        if( NULL == hgRC )
        {
            HRESULT hr = GetLastHr();
            if(SUCCEEDED(hr)) hr = E_FAIL;  //< because LockResource documentation does not say "To get error information call GetLastError."
            return hr;
        }

        if( NULL != m_tex )
            m_tex.Release();

//		The D3DX function is surely much better then my own one 'LoadDDS',
//		however it's the only function from 'D3DX9_40.dll' I need for this project, and the DLL is more then 4MB.

//        return D3DXCreateTextureFromFileInMemory(pDevice, pcvRC, cbResourceSize, &m_tex );
		return LoadDDS( pDevice, pcvRC, cbResourceSize, &m_tex );

    }

    // Release the texture
    inline void Release()
    {
        ASSERT( NULL != m_tex, "CTexture::Release is called for uninitialized texture." );
        ReleaseComPtr( m_tex );
    }

    // Assigns the texture to a stage for the device, and setup sampler state.
    inline HRESULT SetTexture( DWORD Sampler = 0 )
    {
        // Get the device
        CComPtr<IDirect3DDevice9> pDevice;
        HRESULT hr = S_OK;
        ASSERT( NULL != m_tex, "CTexture::SetTexture is called for uninitialized texture." );
        hr = m_tex->GetDevice( &pDevice );  //< Hope the IDirect3DDevice9::AddRef() is called internally?
        if( FAILED(hr) ) return hr;

        // Set texture and mip.filter
        hr = pDevice->SetTexture( Sampler, m_tex );
        if( FAILED(hr) ) return hr;

        const struct samplerState
        {
            D3DSAMPLERSTATETYPE Type;
            DWORD Value;
        }
        arrState[] = 
        {
            D3DSAMP_MIPFILTER, m_texMipFilter,
            D3DSAMP_MINFILTER, m_texMipFilter,
            D3DSAMP_MAGFILTER, m_texMipFilter,
            D3DSAMP_ADDRESSU, D3DTADDRESS_BORDER,
            D3DSAMP_ADDRESSV, D3DTADDRESS_BORDER,
            D3DSAMP_ADDRESSW, D3DTADDRESS_BORDER,
            D3DSAMP_BORDERCOLOR, D3DCOLOR_ARGB(0, 0, 0, 0)
        };

        for( size_t i=0; i<sizeof(arrState)/sizeof(arrState[0]); i++ )
        {
            samplerState ss = arrState[i];
            hr = pDevice->SetSamplerState( Sampler, ss.Type, ss.Value );
            if( FAILED(hr) ) return hr;
        }

        return S_OK;
    }
private:

	// The DDSURFACEDESC2 contains a field of type LPVOID
	// That's why when building for x64 platform, where the 8==sizeof(void*), the DDSURFACEDESC2 structure goes wrong and no longer corresponds to the DDS file layout
#ifdef _M_X64
	struct DdsSurfaceDesc
	{
		DWORD               dwSize;                 // size of the DDSURFACEDESC structure
		DWORD               dwFlags;                // determines what fields are valid
		DWORD               dwHeight;               // height of surface to be created
		DWORD               dwWidth;                // width of input surface
		union
		{
			LONG            lPitch;                 // distance to start of next line (return value only)
			DWORD           dwLinearSize;           // Formless late-allocated optimized surface size
		} DUMMYUNIONNAMEN(1);
		union
		{
			DWORD           dwBackBufferCount;      // number of back buffers requested
			DWORD           dwDepth;                // the depth if this is a volume texture 
		} DUMMYUNIONNAMEN(5);
		union
		{
			DWORD           dwMipMapCount;          // number of mip-map levels requestde
			// dwZBufferBitDepth removed, use ddpfPixelFormat one instead
			DWORD           dwRefreshRate;          // refresh rate (used when display mode is described)
			DWORD           dwSrcVBHandle;          // The source used in VB::Optimize
		} DUMMYUNIONNAMEN(2);
		DWORD               dwAlphaBitDepth;        // depth of alpha buffer requested
		DWORD               dwReserved;             // reserved
		DWORD               lpSurface;              // pointer to the associated surface memory

		union
		{
			DDCOLORKEY      ddckCKDestOverlay;      // color key for destination overlay use
			DWORD           dwEmptyFaceColor;       // Physical color for empty cubemap faces
		} DUMMYUNIONNAMEN(3);
		DDCOLORKEY          ddckCKDestBlt;          // color key for destination blt use
		DDCOLORKEY          ddckCKSrcOverlay;       // color key for source overlay use
		DDCOLORKEY          ddckCKSrcBlt;           // color key for source blt use
		union
		{
			DDPIXELFORMAT   ddpfPixelFormat;        // pixel format description of the surface
			DWORD           dwFVF;                  // vertex format description of vertex buffers
		} DUMMYUNIONNAMEN(4);
		DDSCAPS2            ddsCaps;                // direct draw surface capabilities
		DWORD               dwTextureStage;         // stage in multitexture cascade
	};
#else
	typedef DDSURFACEDESC2 DdsSurfaceDesc;
#endif

	static HRESULT WINAPI LoadDDS( IDirect3DDevice9 *pDevice, const void* pvSrcData, UINT SrcDataSize, IDirect3DTexture9 ** ppTexture )
	{
		const BYTE* pData = static_cast<const BYTE*>(pvSrcData);
		if( *(const DWORD*)(pData) != MAKEFOURCC( 'D', 'D', 'S', ' ') )
			return HRESULT_FROM_WIN32( ERROR_INVALID_DATA );
		const DdsSurfaceDesc *pSurface = (const DdsSurfaceDesc *)( pData + 4 );

		const DWORD dwFlags = pSurface->dwFlags;

		static const DWORD s_TheOnlySupportedFormat = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT | DDSD_LINEARSIZE | DDSD_MIPMAPCOUNT;
		CHECK( s_TheOnlySupportedFormat == dwFlags, return HRESULT_FROM_WIN32(ERROR_UNSUPPORTED_TYPE), "Unsupported DDS format" );

		CHECK( MAKEFOURCC( 'D', 'X', 'T', '5') == pSurface->ddpfPixelFormat.dwFourCC,
			return HRESULT_FROM_WIN32(ERROR_UNSUPPORTED_TYPE),
			"Unsupported DDS format" );

		CComPtr<IDirect3DTexture9> pTex;
		const D3DFORMAT f = (D3DFORMAT)( pSurface->ddpfPixelFormat.dwFourCC );
		const CSize szTex( pSurface->dwWidth, pSurface->dwHeight );
		ASSERT( 1 == ( szTex.cx ^ ( szTex.cx + 1 ) ) && 1 == ( szTex.cy ^ ( szTex.cy + 1 ) ),
			"Compressed textures must be a power of 2" ) ;
		
		HRESULT hr = pDevice->CreateTexture( szTex.cx, szTex.cy, pSurface->dwMipMapCount, 0, f, D3DPOOL_MANAGED, &pTex, NULL );
		if(FAILED(hr))
		{
			ASSERT( 0, "IDirect3DDevice9::CreateTexture failed\n%S", LPCTSTR(ErrorMessage(hr)) );
			return hr;
		}

		const BYTE* pImage = pData + 4 + pSurface->dwSize;
		size_t cbMipLevelSize =  szTex.cx * szTex.cy;	// 'coz DXT5 compresses 1:4 thus giving exactly 1 byte per pixel
		for( DWORD i=0; i<pSurface->dwMipMapCount; i++ )
		{
#ifdef _DEBUG
			D3DSURFACE_DESC dsd;
			hr = pTex->GetLevelDesc( i, &dsd );
			if(FAILED(hr))
			{
				ASSERT( 0, "IDirect3DTexture9::GetLevelDesc(%i) failed.", int(i) );
				return hr;
			}
			ASSERT( dsd.Width == ( szTex.cx >> i ) && dsd.Height == ( szTex.cy >> i ), "Wrong miplevel size" );
#endif
			D3DLOCKED_RECT lrc;
			hr = pTex->LockRect( i, &lrc, NULL, 0 );
			if(FAILED(hr))
			{
				ASSERT( 0, "IDirect3DTexture9::LockRect(%i) failed.", int(i) );
				return hr;
			}

			CopyMemory( lrc.pBits, pImage, cbMipLevelSize );

			hr = pTex->UnlockRect( i );
			if(FAILED(hr))
			{
				ASSERT( 0, "IDirect3DTexture9::UnlockRect(%i) failed.", int(i) );
				return hr;
			}
			
			pImage += cbMipLevelSize;
			cbMipLevelSize /= 4;
		}
		*ppTexture = pTex.Detach();
		return S_OK;
	}
};