#pragma once

template<typename tBuffElement> class CVertexBufferLock;

// This class wraps a Direct3D9 vertex buffer
// The class tBuffElement must implement "static DWORD GetFVF(void)" returning a combination of 'D3DFVF_*' constants.
template<typename tBuffElement>
class CVertexBuffer
{
	static const size_t s_szElement = sizeof(tBuffElement);
    size_t m_nSizeElts;
    CComPtr<IDirect3DVertexBuffer9> m_pBuffer;
public:
    typedef typename CVertexBufferLock<tBuffElement> LockType;

	CVertexBuffer()
    {
        m_nSizeElts = 0;
    }
	~CVertexBuffer(){}

    inline size_t GetCount() const
    {
        return m_nSizeElts;
    }

	// Create the vertex buffer
	HRESULT Create( IDirect3DDevice9* pDevice, size_t nElements, D3DPOOL Pool = D3DPOOL_MANAGED, DWORD Usage = D3DUSAGE_WRITEONLY )
	{
		__if_exists( tBuffElement::GetVertexDeclaration )
		{
            ASSERT( 0, "Achtung! Untested code." );
            IDirect3DVertexBuffer9* pRes = NULL;
            HRESULT hr = pDevice->CreateVertexBuffer( UINT( nElements * sizeof(tBuffElement) ),
                Usage, 0, Pool, &pRes, NULL );
            if( FAILED( hr ) )
                return hr;
            m_nSizeElts = nElements;
            m_pBuffer.Attach( pRes );
            return S_OK;
		}
		__if_exists( tBuffElement::GetFVF )
		{
			const DWORD dwFXV = tBuffElement::GetFVF();
			IDirect3DVertexBuffer9* pRes = NULL;
			HRESULT hr = pDevice->CreateVertexBuffer( UINT( nElements * sizeof(tBuffElement) ),
				Usage, dwFXV, Pool, &pRes, NULL );
			if( FAILED( hr ) )
				return hr;
			m_nSizeElts = nElements;
			m_pBuffer.Attach( pRes );
			return S_OK;
		}
        CHECK( 0, return HRESULT_FROM_WIN32(ERROR_UNSUPPORTED_TYPE),
            "The tBuffElement contains neither GetVertexDeclaration() nor GetFVF()" );
	}

	// Binds a vertex buffer to a device data stream
	HRESULT SetStreamSource( IDirect3DDevice9* pDevice, UINT nStreamNum )
	{
        __if_exists( tBuffElement::GetVertexDeclaration )
		{
            CComPtr<IDirect3DVertexDeclaration9> pDecl;
            HRESULT hr = tBuffElement::GetVertexDeclaration( &pDecl );
            if(FAILED(hr)) return hr;

            hr = pDevice->SetVertexDeclaration( pDecl );
            if(FAILED(hr)) return hr;

            return pDevice->SetStreamSource( nStreamNum, m_pBuffer, 0, UINT( sizeof( tBuffElement ) ) );
		}

        __if_exists( tBuffElement::GetFVF )
        {
            HRESULT hr = pDevice->SetFVF( tBuffElement::GetFVF() );
            if(FAILED(hr)) return hr;
            return pDevice->SetStreamSource( nStreamNum, m_pBuffer, 0, UINT( sizeof( tBuffElement ) ) );
        }
        ASSERT( 0, "The tBuffElement contains neither GetVertexDeclaration() nor GetFVF()" );
        return HRESULT_FROM_WIN32(ERROR_UNSUPPORTED_TYPE);
	}

	// Destroy the vertex buffer
	HRESULT Destroy()
	{
		if( NULL == m_pBuffer )
			return S_FALSE;

        ReleaseComPtr( m_pBuffer );
		m_nSizeElts = 0;
		return S_OK;
	}

	// Lock the buffer, returning the pointer to the raw data
	HRESULT LockBuffer( tBuffElement** ppData, size_t szFirstElement, size_t szCount )
	{
		CHECK( szFirstElement + szCount <= m_nSizeElts,
			return HRESULT_FROM_WIN32(ERROR_INVALID_INDEX),
			"The range is out of bounds" );
		CHECK( NULL != ppData, return E_POINTER, "NULL pointer passed" );

		UINT OffsetToLock = UINT( szFirstElement * s_szElement );
		UINT SizeToLock = UINT( szCount * s_szElement );
		return m_pBuffer->Lock( OffsetToLock, SizeToLock, (void**)ppData, 0 );
	}

	// Release the buffer
	HRESULT UnlockBuffer()
	{
		return m_pBuffer->Unlock();
	}
};

// This class implements release-when-destructed lock for CVertexBuffer,
// allowing elements to be accessed directly with both 'operator []' and 'GetData()'
template<typename tBuffElement>
class CVertexBufferLock
{
public:
	// __ctor: lock a portion of the buffer
	inline CVertexBufferLock( CVertexBuffer<tBuffElement>& buff, size_t iFirstEltToLock, size_t nEltsToLock )
	{
        m_nLocked = 0;
        m_pData = NULL;

        CHECK( nEltsToLock !=0, return, "nEltsToLock must be non-0" );
        CHECK( iFirstEltToLock + nEltsToLock <= buff.GetCount(), return, "ERROR_INVALID_INDEX" );

		m_pBuff = &buff;
		const HRESULT hr = m_pBuff->LockBuffer( &m_pData, iFirstEltToLock, nEltsToLock );
        CHECK( SUCCEEDED(hr), return,
            "CVertexBuffer::LockBuffer failed.\n%S", LPCTSTR(ErrorMessage(hr)) );
        m_nLocked = nEltsToLock;
	}

	// __ctor: lock the whole buffer
    inline CVertexBufferLock( CVertexBuffer<tBuffElement>& buff )
	{
        m_nLocked = 0;
        m_pData = NULL;

        CHECK( buff.GetCount() > 0, return, "The buffer is empty" );

        m_pBuff = &buff;
		const HRESULT hr = m_pBuff->LockBuffer( &m_pData, 0, 0 );
        CHECK( SUCCEEDED(hr), return,
            "CVertexBuffer::LockBuffer failed.\n%S", LPCTSTR(ErrorMessage(hr)) );
        m_nLocked = m_pBuff->GetCount();
	}

	// __dtor: release buffer
	inline ~CVertexBufferLock()
	{
        CHECK( m_nLocked>0, return, "CVertexBufferLock::__dtor is called for an unititialized object." );

		m_pData = NULL;
		const HRESULT hr = m_pBuff->UnlockBuffer();
        ASSERT( SUCCEEDED(hr), "CVertexBuffer::UnlockBuffer failed.\n%S", LPCTSTR(ErrorMessage(hr)) );
	}

    //  Return count of the locked elements
    inline size_t GetCount(){ return m_nLocked; }

	//  Individual elements access
    inline const tBuffElement& operator[]( size_t i ) const
    {
        ASSERT( i < m_nLocked, "ERROR_INVALID_INDEX" );
        return m_pData[i];
    };

	inline tBuffElement& operator[]( size_t i )
    {
        ASSERT( i < m_nLocked, "ERROR_INVALID_INDEX" );
        return m_pData[i];
    };

	// Bulk elements access
    inline tBuffElement* GetData(){ return m_pData; }
private:
    // Locked elements count
	size_t m_nLocked;

    // Pointer to the first locked element
	tBuffElement* m_pData;

    // Owner vertex buffer
	CVertexBuffer<tBuffElement> *m_pBuff;
};