#pragma once
#include "Quaternion.hpp"
#include "Vector3.hpp"

template<typename BaseType>
class Matrix4;

template<typename BaseType>
class __declspec(novtable) Matrix3
{
public:
	typedef Vector3<BaseType> VecType;
	friend Matrix4<BaseType>;
	friend Quaternion<BaseType>& Quaternion<BaseType>::RotateToZY( Vector3<BaseType> newZAxis, Vector3<BaseType> newYAxis );

	// Basic
	inline Matrix3()
	{
		ZeroMemory( m, sizeof(m) );
		m[0][0] = m[1][1] = m[2][2] = BaseType( 1.0 );
	}
	inline Matrix3( const BaseType _m[3][3] )
	{
		CopyFrom( _m );
	}
	inline Matrix3( const Matrix3& that )
	{
		CopyFrom( that.m );
	}

	inline Matrix3& operator=( const Matrix3& that )
	{
		CopyFrom( that.m );
		return *this;
	}
	inline void CopyFrom( const BaseType _m[3][3] )
	{
		CopyMemory( m, _m, sizeof(m) );
	}

	// DirectX-specific
	inline operator D3DMATRIX() const
	{
		D3DMATRIX res;
		for( size_t i=0; i<3; i++ )
			for( size_t j=0; j<3; j++ )
				res.m[i][j] = float( m[i][j] );
		res.m[0][3] = res.m[1][3] = res.m[2][3] = 0;
		res.m[3][0] = res.m[3][1] = res.m[3][2] = 0;
		res.m[3][3] = 1.0f;
		return res;
	}

	// Construct rotation matrix from quaternion
	inline Matrix3( const Quaternion<BaseType> &q )
	{
		// calculate coefficients
		const BaseType x2 = q.x + q.x;
		const BaseType y2 = q.y + q.y;
		const BaseType z2 = q.z + q.z;
		const BaseType xx = q.x * x2;
		const BaseType xy = q.x * y2;
		const BaseType xz = q.x * z2;
		const BaseType yy = q.y * y2;
		const BaseType yz = q.y * z2;
		const BaseType zz = q.z * z2;
		const BaseType wx = q.w * x2;
		const BaseType wy = q.w * y2;
		const BaseType wz = q.w * z2;

		m[0][0] = BaseType(1.0) - (yy + zz);
		m[1][0] = xy - wz;
		m[2][0] = xz + wy;

		m[0][1] = xy + wz;
		m[1][1] = BaseType(1.0) - (xx + zz);
		m[2][1] = yz - wx;

		m[0][2] = xz - wy;
		m[1][2] = yz + wx;
		m[2][2] = BaseType(1.0) - (xx + yy);
	}

	// Construct quaternion from rotation matrix
	inline operator Quaternion<BaseType>() const
	{
		BaseType fTrace = m[0][0] + m[1][1] + m[2][2];
		BaseType fRoot;
		Quaternion<BaseType> res;

		if ( fTrace > 0.0 )
		{
			// |w| > 1/2, may as well choose w > 1/2
			fRoot = sqrt( fTrace + BaseType(1.0) );  // 2w
			res.w = BaseType(0.5) * fRoot;
			fRoot = BaseType(0.5) / fRoot;  // 1/(4w)
			res.x = ( m[1][2] - m[2][1] ) * fRoot;
			res.y = ( m[2][0] - m[0][2] ) * fRoot;
			res.z = ( m[0][1] - m[1][0] ) * fRoot;
		}
		else
		{
			// |w| <= 1/2
			static const size_t s_iNext[3] = { 1, 2, 0 };
			size_t i = 0;
			if ( m[1][1] > m[0][0] )
				i = 1;
			if ( m[2][2] > m[i][i] )
				i = 2;
			size_t j = s_iNext[i];
			size_t k = s_iNext[j];

			fRoot = sqrt( m[i][i] - m[j][j] - m[k][k] + 1.0 );
			BaseType* apkQuat[3] = { &res.x, &res.y, &res.z };
			*apkQuat[i] = BaseType(0.5) * fRoot;
			fRoot = BaseType(0.5) / fRoot;
			res.w = ( m[j][k] - m[k][j] ) * fRoot;
			*apkQuat[j] = ( m[i][j] + m[j][i] ) * fRoot;
			*apkQuat[k] = ( m[i][k] + m[k][i] ) * fRoot;
		}
		return res;
/*
		if ( fTrace > 0.0 )
		{
			// |w| > 1/2, may as well choose w > 1/2
			fRoot = sqrt( fTrace + BaseType(1.0) );  // 2w
			res.w = BaseType(0.5) * fRoot;
			fRoot = BaseType(0.5) / fRoot;  // 1/(4w)
			res.x = ( m[2][1] - m[1][2] ) * fRoot;
			res.y = ( m[0][2] - m[2][0] ) * fRoot;
			res.z = ( m[1][0] - m[0][1] ) * fRoot;
		}
		else
		{
			// |w| <= 1/2
			static const size_t s_iNext[3] = { 1, 2, 0 };
			size_t i = 0;
			if ( m[1][1] > m[0][0] )
				i = 1;
			if ( m[2][2] > m[i][i] )
				i = 2;
			size_t j = s_iNext[i];
			size_t k = s_iNext[j];

			fRoot = sqrt( m[i][i] - m[j][j] - m[k][k] + 1.0 );
			BaseType* apkQuat[3] = { &res.x, &res.y, &res.z };
			*apkQuat[i] = BaseType(0.5) * fRoot;
			fRoot = BaseType(0.5) / fRoot;
			res.w = ( m[k][j] - m[j][k] ) * fRoot;
			*apkQuat[j] = ( m[j][i] + m[i][j] ) * fRoot;
			*apkQuat[k] = ( m[k][i] + m[i][k] ) * fRoot;
		}
		return res;
*/
	}

	// Construct scale matrix from vector
	// THe functionality is not implemented as a constructor because a Vector3 may mean scale or translation,
	// while the quaternions are used exclusively for rotation
	static inline Matrix3 MakeScale( const VecType vScale )
	{
		Matrix3 res;
		res.m[0][0] = vScale.x;
		res.m[1][0] = res.m[2][0] = 0;

		res.m[1][1] = vScale.y;
		res.m[0][1] = res.m[2][1] = 0;

		res.m[2][2] = vScale.z;
		res.m[0][2] = res.m[1][2] = 0;
		return res;
	}

	// Multiply
	inline Matrix3 operator * ( const Matrix3& that ) const
	{
		BaseType res[3][3];
		for (int row = 0; row < 3; row++)
		{
			for (int col = 0; col < 3; col++)
			{
				res[row][col] = 
					m[row][0] * that.m[0][col] +
					m[row][1] * that.m[1][col] +
					m[row][2] * that.m[2][col];
			}
		}
		return Matrix3(res);
	}
	inline Matrix3& operator *= (const Matrix3& that) const
	{
		return operator =( (*this) * that );
	}

	inline VecType ApplyToVector( const VecType& v ) const
	{
		return VecType
			(
			v.x * m[0][0] + v.y * m[1][0] + v.z * m[2][0],
			v.x * m[0][1] + v.y * m[1][1] + v.z * m[2][1],
			v.x * m[0][2] + v.y * m[1][2] + v.z * m[2][2]
		);
	}

	inline const VecType& GetRow( int i ) const
	{
		return *static_cast<const VecType*>( static_cast<const void*>( &m[i][0] ) );
	}
	inline VecType& GetRow( int i )
	{
		return *static_cast<VecType*>( static_cast<void*>( &m[i][0] ) );
	}

	inline VecType GetCol( int i ) const
	{
		return VecType( m[0][i], m[1][i], m[2][i] );
	}

private:
	BaseType m[3][3];
};

template<typename BaseType>
inline Vector3<BaseType> Quaternion<BaseType>::RotateVector( const Vector3<BaseType>& v ) const
{
	Matrix3<BaseType> m(*this);
	return m.ApplyToVector(v);
}

template<typename BaseType>
inline void Quaternion<BaseType>::GetRotatedAxes( Vector3<BaseType> v[3] ) const
{
	Matrix3<BaseType> m(*this);

	v[0] = m.GetCol(0);
	v[1] = m.GetCol(1);
	v[2] = m.GetCol(2);
/*
	v[0] = m.ApplyToVector( Vector3<BaseType>::axisX );
	v[1] = m.ApplyToVector( Vector3<BaseType>::axisY );
	v[2] = m.ApplyToVector( Vector3<BaseType>::axisZ );
*/
}

template<typename BaseType>
inline Quaternion<BaseType>& Quaternion<BaseType>::RotateToZY( Vector3<BaseType> newZAxis, Vector3<BaseType> newYAxis )
{
	// Construct all 3 axis, using normalization and cross product
	newZAxis.Normalize();
	const Vector3<BaseType> newXAxis = newYAxis.Cross(newZAxis).Normalize();	
	newYAxis = newZAxis.Cross( newXAxis ).Normalize();

	// Make the rotation matrix
	Matrix3<BaseType> matRotate;

	matRotate.m[0][0] = newXAxis.x;
	matRotate.m[1][0] = newXAxis.y;
	matRotate.m[2][0] = newXAxis.z;

	matRotate.m[0][1] = newYAxis.x;
	matRotate.m[1][1] = newYAxis.y;
	matRotate.m[2][1] = newYAxis.z;

	matRotate.m[0][2] = newZAxis.x;
	matRotate.m[1][2] = newZAxis.y;
	matRotate.m[2][2] = newZAxis.z;

	// Construct a quaternion from the matrix
	return operator=( matRotate );

	Quaternion<BaseType>& res = operator=( matRotate );
//	Vector3<BaseType> vAxis[3];
//	res.GetRotatedAxes( vAxis );
//	__asm int 3;
	return res;
}

typedef Matrix3<float> Matrix3F;
typedef Matrix3<double> Matrix3D;
