#pragma once
#include "Matrix3.hpp"

namespace UnitTests
{
	// Bitwise compare
	// Calculate difference, then compare length^2 with numeric_limits<T>::epsilon()
	template<typename T>
	bool eq( const Matrix4<T>& a, const Matrix4<T>& b );
}

template<typename BaseType>
class __declspec(novtable) Matrix4
{
public:
	friend bool UnitTests::eq<BaseType>( const Matrix4<BaseType>& a, const Matrix4<BaseType>& b );
	typedef Vector3<BaseType> VecType;

	inline Matrix4& SetIdentity()
	{
		ZeroMemory( m, sizeof(m) );
		m[0][0] = m[1][1] = m[2][2] = m[3][3] = BaseType( 1.0 );
		return *this;
	}

    static inline Matrix4 MakeIdentity()
    {
        Matrix4 res;
        res.SetIdentity();
        return res;
    }

	inline Matrix4()
	{
		SetIdentity();
	}

	inline Matrix4( const BaseType _m[4][4] )
	{
		CopyFrom4(_m);
	}
	inline Matrix4( const BaseType _m[3][3] )
	{
		CopyFrom3(_m);
	}

	inline Matrix4( const Matrix4& that )
	{
		CopyFrom4(that.m);
	}
	inline Matrix4( const Matrix3<BaseType>& that )
	{
		CopyFrom3(that.m);
	}

	inline Matrix4& operator=( const Matrix4& that )
	{
		CopyFrom4(that.m);
		return *this;
	}
	inline Matrix4& operator=( const Matrix3<BaseType>& that )
	{
		CopyFrom3(that.m);
		return *this;
	}

	inline void CopyFrom4( const BaseType _m[4][4] )
	{
		CopyMemory( m, _m, sizeof(m) );
	}
	inline void CopyFrom3( const BaseType _m3[3][3] )
	{
		m[0][0] = _m3[0][0]; m[1][0] = _m3[1][0]; 
		m[2][0] = _m3[2][0]; m[3][0] = 0;

		m[0][1] = _m3[0][1]; m[1][1] = _m3[1][1]; 
		m[2][1] = _m3[2][1]; m[3][1] = 0;

		m[0][2] = _m3[0][2]; m[1][2] = _m3[1][2]; 
		m[2][2] = _m3[2][2]; m[3][2] = 0;

		m[0][3] = 0; m[1][3] = 0;
		m[2][3] = 0; m[3][3] = BaseType(1.0);
	}

	inline operator D3DMATRIX() const;

	inline Matrix4 operator * ( const Matrix4& m2 ) const
	{
		// Copy-pasted from Ogre::Matrix4::concatenate
		BaseType r[4][4];

		r[0][0] = m[0][0] * m2.m[0][0] + m[0][1] * m2.m[1][0] + m[0][2] * m2.m[2][0] + m[0][3] * m2.m[3][0];
		r[0][1] = m[0][0] * m2.m[0][1] + m[0][1] * m2.m[1][1] + m[0][2] * m2.m[2][1] + m[0][3] * m2.m[3][1];
		r[0][2] = m[0][0] * m2.m[0][2] + m[0][1] * m2.m[1][2] + m[0][2] * m2.m[2][2] + m[0][3] * m2.m[3][2];
		r[0][3] = m[0][0] * m2.m[0][3] + m[0][1] * m2.m[1][3] + m[0][2] * m2.m[2][3] + m[0][3] * m2.m[3][3];

		r[1][0] = m[1][0] * m2.m[0][0] + m[1][1] * m2.m[1][0] + m[1][2] * m2.m[2][0] + m[1][3] * m2.m[3][0];
		r[1][1] = m[1][0] * m2.m[0][1] + m[1][1] * m2.m[1][1] + m[1][2] * m2.m[2][1] + m[1][3] * m2.m[3][1];
		r[1][2] = m[1][0] * m2.m[0][2] + m[1][1] * m2.m[1][2] + m[1][2] * m2.m[2][2] + m[1][3] * m2.m[3][2];
		r[1][3] = m[1][0] * m2.m[0][3] + m[1][1] * m2.m[1][3] + m[1][2] * m2.m[2][3] + m[1][3] * m2.m[3][3];

		r[2][0] = m[2][0] * m2.m[0][0] + m[2][1] * m2.m[1][0] + m[2][2] * m2.m[2][0] + m[2][3] * m2.m[3][0];
		r[2][1] = m[2][0] * m2.m[0][1] + m[2][1] * m2.m[1][1] + m[2][2] * m2.m[2][1] + m[2][3] * m2.m[3][1];
		r[2][2] = m[2][0] * m2.m[0][2] + m[2][1] * m2.m[1][2] + m[2][2] * m2.m[2][2] + m[2][3] * m2.m[3][2];
		r[2][3] = m[2][0] * m2.m[0][3] + m[2][1] * m2.m[1][3] + m[2][2] * m2.m[2][3] + m[2][3] * m2.m[3][3];

		r[3][0] = m[3][0] * m2.m[0][0] + m[3][1] * m2.m[1][0] + m[3][2] * m2.m[2][0] + m[3][3] * m2.m[3][0];
		r[3][1] = m[3][0] * m2.m[0][1] + m[3][1] * m2.m[1][1] + m[3][2] * m2.m[2][1] + m[3][3] * m2.m[3][1];
		r[3][2] = m[3][0] * m2.m[0][2] + m[3][1] * m2.m[1][2] + m[3][2] * m2.m[2][2] + m[3][3] * m2.m[3][2];
		r[3][3] = m[3][0] * m2.m[0][3] + m[3][1] * m2.m[1][3] + m[3][2] * m2.m[2][3] + m[3][3] * m2.m[3][3];

		return r;
	}
	inline Matrix4& operator *= (const Matrix4& that)
	{
		return operator =( (*this) * that );
	}


	inline VecType ApplyToVector( const VecType& v ) const
	{
		VecType r;

		const BaseType fInvW = BaseType( 1.0 ) / ( m[3][0] * v.x + m[3][1] * v.y + m[3][2] * v.z + m[3][3] );
		r.x = ( m[0][0] * v.x + m[0][1] * v.y + m[0][2] * v.z + m[0][3] ) * fInvW;
		r.y = ( m[1][0] * v.x + m[1][1] * v.y + m[1][2] * v.z + m[1][3] ) * fInvW;
		r.z = ( m[2][0] * v.x + m[2][1] * v.y + m[2][2] * v.z + m[2][3] ) * fInvW;

		return r;
	}

	inline Matrix4& SetTransition( const VecType& v )
	{
		SetIdentity();
		m[3][0] = v.x;
		m[3][1] = v.y;
		m[3][2] = v.z;
		return *this;
	}
	static inline Matrix4 MakeTransition( const VecType& v )
	{
		Matrix4 res;
		res.SetTransition(v);
		return res;
	}

	HRESULT SetPerspectiveFovLH( BaseType radFovY, BaseType Aspect, BaseType zNear, BaseType zFar )
	{
		*this = Matrix4();
		CHECK( radFovY > BaseType(0.01) && radFovY < BaseType(3.1315), return E_INVALIDARG, "Wrong FOV value %g rad", double(radFovY) );
		CHECK( Aspect > BaseType(0.001) && Aspect < 1000, return E_INVALIDARG, "Wrong Aspect value %g", double(Aspect) );
		CHECK( zNear >  BaseType(1E-4), return E_INVALIDARG, "Too small zNear value %g", double(zNear) );
		CHECK( zFar > zNear, return E_INVALIDARG, "The zFar value must be more then zNear" );

		// Source: D3DXMatrixPerspectiveFovLH documentation
		const BaseType yScale = BaseType( 1.0 ) / tan( radFovY * BaseType(0.5) );
		const BaseType xScale = yScale / Aspect;
		const BaseType zDiff = zFar - zNear;
		m[0][0] = xScale;
		m[1][1] = yScale;
		m[2][2] = zFar / zDiff;

		m[3][2] = -zNear * zFar / zDiff;
		m[2][3] = 1;

		m[3][3] = 0;
		return S_OK;
	}

    HRESULT SetOrthoLH( BaseType w, BaseType h, BaseType zNear, BaseType zFar )
    {
        *this = Matrix4();
        CHECK( w > BaseType(0.01) && h > BaseType(0.01), return E_INVALIDARG, "Width and/or height is either too small or negative or both" );
        CHECK( zNear + BaseType(1E-4) < zFar, return E_INVALIDARG, "The zFar value must be more then zNear" );
        m[0][0] = BaseType(2) / w;
        m[1][1] = BaseType(2) / h;
        m[2][2] = BaseType(1) / ( zFar - zNear );
        m[3][2] = zNear / ( zNear - zFar );
        m[3][3] = BaseType(1);
        return S_OK;
    }
private:
	BaseType m[4][4];
};

template<>
inline Matrix4<float>::operator D3DMATRIX() const
{
	return *static_cast< const D3DMATRIX* >( static_cast< const void* >( this ) );	
}

template<typename BaseType>
inline Matrix4<BaseType>::operator D3DMATRIX() const
{
	D3DMATRIX res;
	const BaseType* pSource = &m[0][0];
	float* pDest = &res.m[0][0];
	for( size_t i=0; i < 16; i++ )
		pDest[i] = float( pSource[i] );
	return res;
}

typedef Matrix4<float> Matrix4F;
typedef Matrix4<double> Matrix4D;
