#pragma once
#include "Vector3.hpp"

// This template class represents a quaternion.
// Keep in mind that integer quaternions are completely useless.

template<typename T>
class Matrix3;

// TODO: implement conversation to/from Euler angles for debugging
template<typename BaseType>
class __declspec(novtable) Quaternion
{
	// Data members; w is a real part (see MSDN documentation on quaternions for the reason why.
	BaseType x, y, z, w;

public:
	friend class Matrix3<BaseType>;
	typedef Vector3<BaseType> VecType;

	// Pi constant is implemented in hardware since 1980
	static inline double GetPi()
	{
#ifndef _M_X64
		__asm fldpi;
#else
		return acos(-1.0);
#endif
	}

	inline Quaternion()
	{
		w = 1;
		x = y = z = 0;
	}
	inline void CopyFrom( const Quaternion& that )
	{
		x = that.x;
		y = that.y;
		z = that.z;
		w = that.w;
	}
	inline Quaternion( const Quaternion& that ) { CopyFrom(that); }
	inline Quaternion& operator=( const Quaternion& that ) { CopyFrom(that); return *this; }

	// Change the type.
	template<class OtherType>
	inline Quaternion<OtherType> ChangeType()
	{
		Quaternion<OtherType> res;
		res.x = OtherType(x);
		res.y = OtherType(y);
		res.z = OtherType(z);
		res.w = OtherType(w);
		return res;
	}

//	Multiplication
//	Source: http://www.gamasutra.com/view/feature/3278/rotating_objects_using_quaternions.php?page=2
	inline Quaternion& operator*=( const Quaternion& that )
	{
		const BaseType A = ( w + x ) * ( that.w + that.x );
		const BaseType B = ( z - y ) * ( that.y - that.z );
		const BaseType C = ( w - x ) * ( that.y + that.z );
		const BaseType D = ( y + z ) * ( that.w - that.x );
		const BaseType E = ( x + z ) * ( that.x + that.y );
		const BaseType F = ( x - z ) * ( that.x - that.y );
		const BaseType G = ( w + y ) * ( that.w - that.z );
		const BaseType H = ( w - y ) * ( that.w + that.z );

		Quaternion res;
		res.w = B + ( -E - F + G + H ) / 2;
		res.x = A - (  E + F + G + H ) / 2;
		res.y = C + (  E - F + G - H ) / 2; 
		res.z = D + (  E - F - G + H ) / 2;
		CopyFrom( res );
		return *this;
	}

	// Spherical linear interpolation: return this*(1.0 - blendFactor) + blendWith*blendFactor
	// Source: http://www.gamasutra.com/view/feature/3278/rotating_objects_using_quaternions.php?page=2
	Quaternion Interpolate( const Quaternion &blendWith, BaseType blendFactor ) const
	{
		CHECK( blendFactor>=0, blendFactor=0, "Interpolate: the factor is negative" );
		CHECK( blendFactor<=1.0, blendFactor=1.0, "Interpolate: the factor is more then 1" );

		static const BaseType DELTA = 1E-4;

		BaseType to1[4];
		double omega, cosom, sinom, scale0, scale1;

		// calculate cosine
		cosom = x * blendWith.x + y * blendWith.y + z * blendWith.z
			+ w * blendWith.w;

		// adjust signs (if necessary)
		if ( cosom < 0.0 )
		{
			cosom = -cosom;
			to1[0] = -blendWith.x;
			to1[1] = -blendWith.y;
			to1[2] = -blendWith.z;
			to1[3] = -blendWith.w;
		}
		else 
		{
			to1[0] = blendWith.x;
			to1[1] = blendWith.y;
			to1[2] = blendWith.z;
			to1[3] = blendWith.w;
		}

		// Calculate the coefficients
		if ( ( 1.0 - cosom ) > DELTA )
		{
			// standard case - SLERP
			omega = acos(cosom);
			sinom = sin(omega);
			scale0 = sin((1.0 - blendFactor) * omega) / sinom;
			scale1 = sin(blendFactor * omega) / sinom;
		}
		else
		{
			// "this" and "blendWith" quaternions are very close,
			// so we can do a linear interpolation
			scale0 = 1.0 - blendFactor;
			scale1 = blendFactor;
		}
		// calculate final values
		Quaternion res;
		res.x = scale0 * x + scale1 * to1[0];
		res.y = scale0 * y + scale1 * to1[1];
		res.z = scale0 * z + scale1 * to1[2];
		res.w = scale0 * w + scale1 * to1[3];
		res.Normalize();
		return res;
	}

	// TODO: normalize names :-)
	// rotation from axis-angle
	inline static Quaternion ConstructRotation( VecType axis, BaseType radAngle )
	{
		radAngle /= 2;
		axis.Normalize();
		const BaseType sinAngle = sin( radAngle );
		Quaternion res;
		res.x = axis.x * sinAngle;
		res.y = axis.y * sinAngle;
		res.z = axis.z * sinAngle;
		res.w = cos( radAngle );
		return res;
	}
	inline Quaternion& RotateByAxis( VecType axis, BaseType radAngle )
	{
		const Quaternion q = ConstructRotation( axis, radAngle );
		return operator *=( q );
	}

	// Rotate a vector with this quaternion
	// see Matrix3.hpp for implementation
	inline VecType RotateVector( const VecType& v ) const;

	// Get positive XYZ axes altered by the rotation
	inline void GetRotatedAxes( VecType vAxes[3] ) const;

	// Construct a rotation so that transformed Z direction is newZAxis, and new Y direction is newYAxis
	Quaternion& RotateToZY( VecType newZAxis, VecType newYAxis );

	Quaternion Inverse() const
	{
		Quaternion res(*this);
		res.x = -res.x;
		res.y = -res.y;
		res.z = -res.z;
		return res;
	}
private:
	void Normalize()
	{
		const BaseType scale = BaseType(1) / sqrt( x*x + y*y + z*z + w*w );
		x *= scale;
		y *= scale;
		z *= scale;
		w *= scale;
	}
};

template <typename T>
const Quaternion<T> operator*( const Quaternion<T>& lhs, const Quaternion<T>& rhs )
{
	return Quaternion<T>(lhs) *= rhs;
}

typedef Quaternion<float> QuaternionF;
typedef Quaternion<double> QuaternionD;

inline double radFromDeg( double deg )
{
	return deg * QuaternionD::GetPi() / 180.0;
}

inline float radFromDeg( float deg )
{
	return deg * float( QuaternionD::GetPi() ) / 180.0f;
}
