#include "stdafx.h"
#include "Transform.hpp"

namespace UnitTests
{
	template<class T>
	HRESULT vectorsBasicOperations()
	{
		typedef Vector3<T> VecType;
		// default constructor
		const VecType v0;
		if( v0.x != 0 || v0.y != 0 || v0.z != 0 )
			__debugbreak();

		// operator + and non-default constructor
		const VecType v1( 1, 2, 3 );
		if( v0 + v1 != v1 )
			__debugbreak();

		// operator -
		if( (-v1) != v0 - v1 )
			__debugbreak();

		if( v1 * 2 != VecType( 2, 4, 6 ) )
			__debugbreak();

		return S_OK;
	}
	UNIT_TEST( vectorsBasicOperations<float> );
	UNIT_TEST( vectorsBasicOperations<double> );

	template<class T>
	HRESULT QuaternionsRotations()
	{
		typedef Vector3<T> VecType;
		typedef Quaternion<T> QuatType;

		const VecType axisX( 1, 0, 0 );
		const VecType axisY( 0, 1, 0 );
		const VecType axisZ( 0, 0, 1 );

		const VecType vTest( 1, -2, 3 );

		const QuatType q0;
		if( vTest != q0.RotateVector( vTest ) )
			__debugbreak();

		// Rotate 180 around the Z
		const QuatType q1 = QuatType(q0).RotateByAxis( axisZ, radFromDeg( T(180) ) );
		const VecType vTest1 = q1.RotateVector( vTest );
		if( !eq( vTest1, VecType( -1, 2, 3 ) ) )
			__debugbreak();

		// Rotate once more
		const VecType vTest2 = q1.RotateVector( vTest1 );
		if( !eq( vTest2, vTest ) )
			__debugbreak();

		// Do the same, but rotate around the arbitrary axis, and multiply the quaternions
		const VecType vAxis(-2.68687354f, 1.354987f, -1.2654684f);
		QuatType q3 = q0;
		q3.RotateByAxis( vAxis, radFromDeg( T(34) ) );
		q3.RotateByAxis( vAxis, radFromDeg( T(-198) ) );
		q3.RotateByAxis( vAxis, radFromDeg( T(198 - 34) ) );
		const VecType vTest3 = q3.RotateVector( vTest );
		if( !eq( vTest3, vTest ) )
			__debugbreak();

		QuatType q4 = q0;
		q4.RotateByAxis( axisZ, radFromDeg( T(90) ) );
		VecType vRotatedAxes[3];
		q4.GetRotatedAxes(vRotatedAxes);
		if(! eq( vRotatedAxes[2], axisZ ) )
			__debugbreak();
		if(! eq( vRotatedAxes[1], axisX ) )
			__debugbreak();
		if(! eq( vRotatedAxes[0], -axisY ) )
			__debugbreak();

		return S_OK;
	}
	UNIT_TEST( QuaternionsRotations<float> );
	UNIT_TEST( QuaternionsRotations<double> );

	template<class T>
	HRESULT Matrix3Ops()
	{
		typedef Vector3<T> VecType;
		typedef Quaternion<T> QuatType;
		typedef Matrix3<T> MatType;

		const VecType vAxis(-2.68687354f, 1.354987f, -1.2654684f);
		const MatType r1( QuatType::ConstructRotation( vAxis, T(34) ) );
		const MatType r2( QuatType::ConstructRotation( vAxis, T(-198) ) );
		const MatType r3( QuatType::ConstructRotation( vAxis, T(198-34) ) );
		const MatType r = r1 * r2 * r3;

		const VecType vToRotate(1.25f, 2.9387f, 3.981723f);
		const VecType vRotated = r.ApplyToVector(vToRotate);
		if( !eq( vToRotate, vRotated ) )
			__debugbreak();
		
		return S_OK;
	}
	UNIT_TEST( Matrix3Ops<float> );
	UNIT_TEST( Matrix3Ops<double> );

	template<class T>
	HRESULT Transforms()
	{
		typedef Vector3<T> VecType;
		typedef Quaternion<T> QuatType;
		typedef Matrix4<T> MatType;
		typedef Transform<T> TransType;

		TransType t0;
		const VecType vTest(1.25f, 2.9387f, 3.981723f);
		const VecType vTest_t = MatType(t0).ApplyToVector(vTest);
		if( !eq( vTest, vTest_t ) )
			__debugbreak();

		return S_OK;
	}
	UNIT_TEST( Transforms<float> );
	UNIT_TEST( Transforms<double> );

	template<class T>
	inline bool eq( const Matrix4<T>& a, const Matrix4<T>& b )
	{
		const T eps = std::numeric_limits<T>::epsilon();
		const T eps2 = sqrt(eps);
		for( size_t i=0; i<4; i++ )
			for( size_t j=0; j<4; j++ )
			{
				const T absDiff = abs( a.m[i][j] - b.m[i][j] );
				if( absDiff > eps2 )
					return false;
			}
			return true;
	}

#ifdef __D3DX9MATH_H__
	HRESULT MatrixVersusD3dx()
	{
		// Identity
		Matrix4F matMy;
		D3DXMATRIX matD3D;
		D3DXMatrixIdentity( &matD3D );
		if( !eq( matMy, Matrix4F( matD3D.m ) ) )
			__debugbreak();

		// Translation
		const Vector3F vTest( 1.25f, 2.9387f, 3.981723f );
		D3DXMatrixTranslation( &matD3D, vTest.x, vTest.y, vTest.z  );
		matMy.SetTransition( vTest );
		if( !eq( matMy, Matrix4F( matD3D.m ) ) )
			__debugbreak();

		// Perspective projection
		Matrix4F matMy2;
		matMy2.SetPerspectiveFovLH( radFromDeg(90.0f), 1.2f, 1, 10 );
		D3DXMATRIX matD3D2;
		D3DXMatrixPerspectiveFovLH( &matD3D2, radFromDeg(90.0f), 1.2f, 1, 10 );
		if( !eq( matMy2, Matrix4F( matD3D2.m ) ) )
			__debugbreak();

		// Multiplication
		const Matrix4F matMy3 = matMy * matMy2;
		D3DXMATRIX matD3D3;
		D3DXMatrixMultiply( &matD3D3, &matD3D, &matD3D2 );
		if( !eq( matMy3, Matrix4F( matD3D3.m ) ) )
			__debugbreak();

		return S_OK;
	}

	UNIT_TEST( MatrixVersusD3dx );
#endif
}
