#pragma once
#include "Matrix4.hpp"

// This template class implements an affine transform in 3D space
template<typename BaseType>
class __declspec(novtable) Transform
{
public:
	typedef Vector3<BaseType> VecType;
	typedef Quaternion<BaseType> QuatType;

	// Typical C++ stuff
	inline Transform() : m_scale( 1, 1, 1) { }
	inline ~Transform(){}

	inline void CopyFrom( const Transform& that )
	{
		m_position.CopyFrom( that.m_position );
		m_orientation.CopyFrom( that.m_orientation );
	}

	inline Transform( const Transform& that ) { CopyFrom( that ); }
	inline Transform& operator= ( const Transform& that ) { CopyFrom( that ); return *this; }

	// Direct access to the position
	inline const VecType& GetPosition() const { return m_position; }
	inline VecType& GetPosition() { return m_position; }
	inline void SetPosition( const VecType& pos ) { m_position = pos; }
	__declspec ( property( get=GetPosition, put=SetPosition ) ) VecType Position;

	// Direct access to the orientation
	inline const QuatType& GetOrientation() const { return m_orientation; }
	inline QuatType& GetOrientation() { return m_orientation; }
	inline void SetOrientation( const QuatType& or ) { m_orientation = or; }
	__declspec ( property( get=GetOrientation, put=SetOrientation ) ) QuatType Orientation;

	// Direct access to the scale
	inline const VecType& GetScale() const { return m_scale; }
	inline VecType& GetScale() { return m_scale; }
	inline void SetScale( const VecType& scale ) { m_scale = scale; }
	__declspec ( property( get=GetScale, put=SetScale ) ) VecType Scale;

	// Operations
	inline operator Matrix4<BaseType>() const
	{
		const Matrix3<BaseType> mRotate(m_orientation);
		const Matrix3<BaseType> mScale = Matrix3<BaseType>::MakeScale(m_scale);
		Matrix4<BaseType> res = ( mRotate * mScale );
		res *= Matrix4<BaseType>().SetTransition( m_position );
		return res;
//		return mRotate * mShift;
	}

	// everything(positions, orientation, scale) move linearly
	inline Transform Interpolate( const Transform& that, BaseType valThat )
	{
		const BaseType valThis = BaseType( 1.0 ) - valThat;
		Transform res;
		res.m_position = m_position * valThis + that.m_position * valThat;
		res.m_scale = m_scale * valThis + that.m_scale * valThat;
		res.m_orientation = m_orientation.Interpolate( that.m_orientation, valThat );
		return res;
	}
private:
	VecType m_position;
	VecType m_scale;
	QuatType m_orientation;
};

typedef Transform<double> TransformD;