#pragma once

// The class implements a vector in 3D space
// The primary purpose of this is Vector3<float> and Vector3<double>, integer vectors are less supported at the moment.
template< typename BaseType>
class __declspec(novtable) Vector3
{
public:
	// Data members
	BaseType x, y, z;

	// std::numeric_limits derived constants
	static const BaseType valMin;

	// Basic C++ stuff
	inline Vector3(){ x=y=z=0; };
	inline ~Vector3(){}

	inline Vector3( BaseType _x, BaseType _y, BaseType _z )
	{
		x=_x; y=_y; z=_z;
	};
	inline Vector3( const Vector3& that ){ CopyFrom(that); }
	inline Vector3& operator = ( const Vector3& that ){ CopyFrom(that); return *this; }

	inline void CopyFrom( const Vector3& that )
	{
		x = that.x;
		y = that.y;
		z = that.z;
	}

	// Addition and scale
	inline Vector3 operator - () const
	{
		Vector3 res;
		res.x = -x;
		res.y = -y;
		res.z = -z;
		return res;
	}

	inline Vector3& operator += ( const Vector3& that )
	{
		x += that.x;
		y += that.y;
		z += that.z;
		return *this;
	}
	inline Vector3 operator + ( const Vector3& that ) const
	{
		return Vector3(*this) += that;
	}

	inline Vector3& operator -= ( const Vector3& that )
	{
		operator+=(-that);
		return *this;
	}
	inline Vector3 operator - ( const Vector3& that ) const
	{
		return Vector3(*this) -= that;
	}

	inline Vector3& operator *= ( BaseType c )
	{
		x *= c;
		y *= c;
		z *= c;
		return *this;
	}
	inline Vector3 operator * ( BaseType that ) const
	{
		return Vector3(*this) *= that;
	}

	inline Vector3& operator /= ( BaseType c )
	{
		ASSERT( abs(currLength) > valMin, "Vector3::operator/=() called with a zero argument" );
		x /= c;
		y /= c;
		z /= c;
		return *this;
	}
	inline Vector3 operator / ( BaseType rhs ) const
	{
		return Vector3(*this) /= rhs;
	}

	// Change type on the vector.
	// It's implemented as explicit operation,
	// because back and forth float-double conversations are bad both for performance and for precision
	template<class OtherType>
	Vector3<OtherType> ChangeType()
	{
		Vector3<OtherType> res;
		res.x = OtherType(x);
		res.y = OtherType(y);
		res.z = OtherType(z);
		return res;
	}

	// Length stuff
	inline BaseType GetLengthSq() const
	{
		return x*x + y*y + z*z;
	}
	__declspec( property( get=GetLengthSq ) ) BaseType LengthSq;

	inline BaseType GetLength() const
	{
		return sqrt( GetLengthSq() );
	}

	inline void SetLength( BaseType newLength )
	{
		const BaseType currLength = GetLength();
		CHECK( abs(currLength) > valMin, return, "Vector3::SetLength() called for a zero-length vector" );
		operator*=( newLength/currLength );
	}
	__declspec( property( get=GetLength, put=SetLength ) ) BaseType Length;

	inline Vector3 Normalize()
	{
		Length = 1;
		return *this;
	}

	//		Cross product, dot product, and scale.
	//		As you can see, they are 3 completely different types of multiplication.
	//		That's exactly why none of them is declared as 'operator * '

	// dot product, also known as scalar product
	inline BaseType Dot( const Vector3& that ) const
	{
		return x * that.x + y * that.y + z * that.z;
	}

	// cross product, also known as vector product
	inline Vector3 Cross( const Vector3& that ) const
	{
		Vector3 res;
		res.x = y * that.z - z * that.y;
		res.y = z * that.x - x * that.z;
		res.z = x * that.y - y * that.x;
//		return ( s_bRightHanded ) ? res : -res;
		return res;
	}

	// Scale each component by the corresponding component of another vector
	inline Vector3 ScaleBy( const Vector3& that ) const
	{
		return Vector3( x * that.x, y * that.y, z * that.z );
	}

	// Scale each component by the inverse component of another vector
	inline Vector3 InvScaleBy( const Vector3& that ) const
	{
		CHECK( abs(that.x) > valMin && abs(that.y) > valMin && abs(that.z) > valMin,
			return Vector3(),
			"Vector3::InvScaleBy() called with a zero argument" );
		return Vector3( x / that.x, y / that.y, z / that.z );
	}

	inline operator D3DVECTOR() const;

	static const Vector3 &axisX, &axisY, &axisZ;
	static const Vector3 axes[3];
};

template <typename T>
const T Vector3<T>::valMin = std::numeric_limits<T>::min();

template <typename T>
const Vector3<T> Vector3<T>::axes[3] = 
{
	Vector3<T>(1,0,0),
	Vector3<T>(0,1,0),
	Vector3<T>(0,0,1)
};

template <typename T>
const Vector3<T> &Vector3<T>::axisX = Vector3<T>::axes[0];

template <typename T>
const Vector3<T> &Vector3<T>::axisY = Vector3<T>::axes[1];

template <typename T>
const Vector3<T> &Vector3<T>::axisZ = Vector3<T>::axes[2];

// DirectX-specific
template <>
inline Vector3<float>::operator D3DVECTOR() const
{
	return *static_cast< const D3DVECTOR* >( static_cast< const void* >( this ) );
}

template <typename T>
inline Vector3<T>::operator D3DVECTOR() const
{
	D3DVECTOR res;
	res.x = float(x);
	res.y = float(y);
	res.z = float(z);
	return res;
}

// The straightforward floating point comparison should never appear in the production code.
// However it's helpful during the unit tests, hence the dedicated namespace.
namespace UnitTests
{
	// Bitwise compare
	template<class T>
	inline bool operator==( const Vector3<T>& a, const Vector3<T>& b )
	{
		return a.x == b.x && a.y == b.y && a.z == b.z;
	}
	template<class T>
	inline bool operator!=( const Vector3<T>& a, const Vector3<T>& b )
	{
		return !( a==b );
	}

	// Calculate difference, then compare length^2 with numeric_limits<T>::epsilon()
	template<class T>
	inline bool eq( const Vector3<T>& a, const Vector3<T>& b )
	{
		const Vector3<T> diff = a - b;
		const T sq = diff.LengthSq;
		const T eps = std::numeric_limits<T>::epsilon();
		return sq < eps;
	}
}

typedef Vector3<double> Vector3D;
typedef Vector3<float> Vector3F;
