#pragma once

// This template container stores a set of object in the infinite 2D space.
// The 2D space is divided into regular grids cells, being referenced by CPoint.
// Every object belongs to exactly one cell.

// The diabolic container automagically builds and maintains the object's mapping to the cells,
// thus allowing cheap "gimme all objects in a given cell, you stupid computer" operations.

// tValue must implement 'CPoint GetCell() const'.
// tValue is assumed to be small,
// e.g. for simplicity 'Update' function physicalloy relocates the cell, while only a few pointers need to be exchanged.

#define USE_STL_INSTEAD 0

#if USE_STL_INSTEAD
#pragma warning ( disable : 4530 )
#include <list>
#include <hash_map>

template<typename T>
class CCrappyList: public std::list<T>
{
public:
	inline POSITION GetHeadPosition() const
	{
		return makePosition( begin() );
	}

	inline size_t GetCount() const
	{
		return size();
	}

	inline T& GetAt( POSITION pos ) const
	{
		const T& val = *makeIter( pos );
		return const_cast<T&>(val);
	}

	inline void GetNext( POSITION& pos ) const
	{
		const_iterator it = makeIter( pos );
		it++;
		pos = makePosition( it );
	}

	inline POSITION AddTail( const T& element )
	{
		push_back( element );
		const_iterator it = end();
		it--;
		return makePosition( it );
	}

	inline POSITION AddHead( const T& element )
	{
		push_front( element );
		return makePosition( begin() );
	}

	inline POSITION InsertAfter( POSITION pos, const T& element )
	{
		const_iterator it = makeIter( pos );
		it++;
		insert( it, element );
		it--;
		return makePosition( it );
	}

	inline POSITION InsertBefore( POSITION pos, const T& element )
	{
		const_iterator it = makeIter( pos );
		insert( it, element );
		it--;
		return makePosition( it );
	}

	inline void RemoveAt( POSITION pos )
	{
		erase( makeIter( pos ) );
	}

	inline void RemoveAll() { clear(); }
private:
	inline const const_iterator makeIter( POSITION pos ) const
	{
		const _Nodeptr nptr = (_Nodeptr)(pos);
#if _SECURE_SCL
		return const_iterator( nptr, this );
#else
		return const_iterator( nptr );
#endif
	}

	inline POSITION makePosition( const_iterator it ) const
	{
		if(it==end()) return NULL;
		return (POSITION)it._Mynode();
	}
};
/*
struct CrappyHashmapHashConfig
{
	enum {
		bucket_size = 4,
		min_buckets = 8};

	inline size_t operator()( const CPoint& key_value ) const
	{
		return CElementTraits<CPoint>::Hash( key_value );
	}

	bool operator()(const CPoint& left, const CPoint& right) const
	{
		if( left.y < right.y )
			return true;
		if( left.x < right.x )
			return true;
		return false;
	}
};

template< typename K, typename V >
class CCrappyHashmap: public stdext::hash_map< K, V, CrappyHashmapHashConfig >
{
public:
	class CPair : public std::pair< const K, V >
	{
	public:
		inline V& getValue() { return second; }
		inline const V& getValue() const { return second; }
		__declspec(property(get=getValue)) V m_value;
	};

	inline CPair* Lookup( K key ) const
	{
		const_iterator it = find(key);
		if( it==end() ) return NULL;
		const std::pair< const K, V >& res = *it;
		return &const_cast<CPair&>(reinterpret_cast<const CPair&>(res));
	}

	inline size_t GetCount() const
	{
		return size();
	}

	inline void RemoveAll() { clear(); }

	inline void RemoveKey( K key )
	{
		iterator it = find(key);
		if( it==end() ) return;
		erase( it );
	}

	inline POSITION GetStartPosition()
	{
		iterator it = begin();
		return (POSITION)it._Mynode();
	}

	inline CPair* GetNext( POSITION& pos );

	inline const K& GetKeyAt( POSITION pos ) const;

}; */
#endif

template<class tValue>
class CBubbleContainer
{
	static const bool s_bRepaintBubblesToVisualizeCells = true;
//	static const bool s_bRepaintBubblesToVisualizeCells = false;

	struct tValueAndCell
	{
		tValue val;
		CPoint cell;
	};
#if USE_STL_INSTEAD
	typedef CCrappyList< tValueAndCell > tList;
#else
	typedef CAtlList< tValueAndCell > tList;
#endif
	tList m_list;

	// the POSITIONs are in the m_list
	struct tMapValue
	{
		inline tMapValue()
		{
			posFirst = NULL;
			nElts = 0;
		}

		inline tMapValue( POSITION pos )
		{
			posFirst = pos;
			nElts = 1;
		}
		POSITION posFirst;
		size_t nElts;
	};

//#if USE_STL_INSTEAD
//	typedef CCrappyHashmap< CPoint, tMapValue > tMap;
//#else
	typedef CAtlMap< CPoint, tMapValue > tMap;
//#endif
	tMap m_map;

	CPoint m_cellMin, m_cellMax;
public:
	CBubbleContainer()
	{
		m_cellMin = CPoint( std::numeric_limits<LONG>::max(), std::numeric_limits<LONG>::max() );
		m_cellMax = CPoint( std::numeric_limits<LONG>::min(), std::numeric_limits<LONG>::min() );
	}

	// add an element
	HRESULT Add( const tValue& elt )
	{
		const CPoint cell = elt.GetCell();
		tValueAndCell val = { elt, cell };

		// debug code: paint bubbles according to the cell
		if( s_bRepaintBubblesToVisualizeCells )
			// val.val.selfColor = BYTE( ( ( 10000 + cell.x + 3 * cell.y ) % 9 ) * ( 256 / 9 ) );
			val.val.selfColor = RgbFromHsv( BYTE( ( ( 10000 + cell.x + 3 * cell.y ) % 9 ) * ( 256 / 9 ) ), 255, 255 );

		tMap::CPair *p = m_map.Lookup( cell );
		if( NULL != p )
		{
			// There're bubbles in that cell already, so we just add them
			m_list.InsertAfter( p->m_value.posFirst, val );
			p->m_value.nElts++;
			return S_OK;
		}

		// No bubbles in that cell yet.

		if( cell.y < m_cellMin.y || ( cell.y == m_cellMin.y && cell.x < m_cellMin.x ) )
		{
			// this element must become a new head of the list
			m_cellMin.x = std::min( m_cellMin.x, cell.x );
			m_cellMin.y = std::min( m_cellMin.y, cell.y );

			if( 0 == m_map.GetCount() )
				m_cellMax = cell;	// and the new tail, as well.

			POSITION posNew = m_list.AddHead( val );
			m_map[cell] = tMapValue( posNew );
			return S_OK;
		}
		
		if ( cell.y > m_cellMax.y || ( cell.y ==  m_cellMax.y && cell.x > m_cellMax.x ) )
		{
			// this element must become a new tail of the list
			m_cellMax.x = std::max( m_cellMax.x, cell.x );
			m_cellMax.y = std::max( m_cellMax.y, cell.y );
			POSITION posNew = m_list.AddTail( val );
			m_map[cell] = tMapValue( posNew );
			return S_OK;
		}

		// worst case ever: this cell is somewhere else so we must find a correct position

		// Search towards the lower-right corner
		// This strategy works best when objects are relatively dense within the [ m_cellMin - m_cellMax ] rectangle.
		// When the objects are sparse,
		// you better cache the tList's positions of the first element within every y=const cell line (or NULL for lines where there's no elements)
		static const LONG lMax = std::numeric_limits<LONG>::max();
		{
			LONG startX = cell.x + 1;
			for( CPoint i = cell; i.y <= m_cellMax.y; i.y++ )
			{
				for( i.x = startX; i.x <= m_cellMax.x; i.x++ )
				{
					tMap::CPair *p = m_map.Lookup( i );
					if( NULL == p )
						continue;
					// Found!
					const POSITION posNew = m_list.InsertBefore( p->m_value.posFirst, val );
					m_map[ cell ] = tMapValue( posNew );
					return S_OK;
				}
				startX = m_cellMin.x;
			}
		}
		// Not found, which means we should add the element to the end of the list
		const POSITION posNew = m_list.AddTail( val );
		m_map[ cell ] = tMapValue( posNew );
		return S_OK;
	}

	HRESULT Remove( POSITION posList )
	{
		const CPoint cell = m_list.GetAt(posList).cell;
		tMap::CPair* p = m_map.Lookup(cell);
		ASSERT( NULL != p, "Data inconsistency: the bubble being removed was not on the hashmap" );
		ASSERT( 0 != p->m_value.nElts, "Data inconsistency" );
		if( p->m_value.nElts > 1 )
		{
			// There're still bubbles in this cell left
			p->m_value.nElts--;
			if( p->m_value.posFirst == posList )
			{
				m_list.GetNext( p->m_value.posFirst );
				ASSERT( NULL != p->m_value.posFirst, "Data inconsistency" );
			}
			m_list.RemoveAt( posList );
			return S_OK;
		}

		// It was the last element in the cell, so we remove both the bubble and the cell
		if( 1 == m_list.GetCount() )
		{
			// ...and it was also the last bubble ever
			ASSERT( 1 == m_map.GetCount(), "Data inconsistency" );
			m_list.RemoveAll();
			m_map.RemoveAll();
			m_cellMin = CPoint( std::numeric_limits<LONG>::max(), std::numeric_limits<LONG>::max() );
			m_cellMax = CPoint( std::numeric_limits<LONG>::min(), std::numeric_limits<LONG>::min() );
			return S_OK;
		}

		m_list.RemoveAt( posList );
		m_map.RemoveKey( cell );
		if( cell.x == m_cellMin.x || cell.y == m_cellMin.y || cell.x == m_cellMax.x || cell.y == m_cellMax.y )
		{
			// Achtung! The code below is O(N),
			// while checking individual sides of (m_cellMin-m_cellMax) rectangle is more complex to code but O(sqrt(N))
			m_cellMin = CPoint( std::numeric_limits<LONG>::max(), std::numeric_limits<LONG>::max() );
			m_cellMax = CPoint( std::numeric_limits<LONG>::min(), std::numeric_limits<LONG>::min() );

			for( POSITION posMap = m_map.GetStartPosition(); NULL != posMap; m_map.GetNext( posMap ) )
			{
				const CPoint ptKey = m_map.GetKeyAt( posMap );
				m_cellMin.x = std::min( m_cellMin.x, ptKey.x );
				m_cellMin.y = std::min( m_cellMin.y, ptKey.y );
				m_cellMax.x = std::max( m_cellMax.x, ptKey.x );
				m_cellMax.y = std::max( m_cellMax.y, ptKey.y );
			}
		}
		return S_OK;
	}

	HRESULT Update( POSITION pos )
	{
		tValueAndCell& vac = m_list.GetAt(pos);
		const CPoint ptThis = vac.cell;
		const CPoint ptNew = vac.val.GetCell();
		if( ptThis == ptNew )
			return S_FALSE;	// No relocation is required, we're fine.
		const tValue val = vac.val;

		HRESULT hr = Remove( pos );
		if(FAILED(hr)) return hr;
		hr = Add( val );
		if(FAILED(hr)) return hr;
		return S_OK;
	}

	inline CPoint GetMinCell() const { return m_cellMin; } 
	inline CPoint GetMaxCell() const { return m_cellMax; } 

	// Get total count of bubbles.
    inline size_t GetBubblesCount() const
	{
		return m_list.GetCount();
	}

	// Get total count of bubbles, and return the POSITION of the first one.
    inline size_t GetAllBubbles( POSITION& pos ) const
	{
		pos = m_list.GetHeadPosition();
		return m_list.GetCount();
	}

	// Get the count of bubbles in the cell, and return the POSITION of the first one.
    inline size_t GetCellBubbles( CPoint cell,  POSITION& pos ) const
	{
		const tMap::CPair* p = m_map.Lookup( cell );
		if( NULL == p )
		{
			pos = NULL;
			return 0;
		}
		ASSERT( p->m_value.nElts > 0 && NULL != p->m_value.posFirst, "Data inconsistency" );
		pos = p->m_value.posFirst;
		return p->m_value.nElts;
	}

	// Get current value, and advance the POSITION
    inline const tValue& GetValueAndNextPos( POSITION& pos ) const
	{
		const tValue& ret = m_list.GetAt( pos ).val;
		m_list.GetNext( pos );
		return ret;
	}

    // Get current value, and advance the POSITION
	inline tValue& GetValueAndNextPos( POSITION& pos )
	{
		tValue& ret = m_list.GetAt( pos ).val;
		m_list.GetNext( pos );
		return ret;
	}
};