#include "stdafx.h"
#include "SoapBubbles.h"

void CSoapBubblesSim::ProcessInteractionsWithinBuffer( typeBubbleProc proc, tTmpBuffer& buff ) const
{
    for( size_t a = 0; a < buff.GetCount(); a++ )
    {
        BubbleSim &ea = *buff.GetAt( a );
        for( size_t i = 0; i < a; i++ )
        {
            BubbleSim &ei = *buff.GetAt( i );
            ( this->*proc ) ( ea, ei );
        }
    }
}

void CSoapBubblesSim::ProcessInteractionsAcrossBuffers( typeBubbleProc proc, tTmpBuffer& b1, tTmpBuffer& b2 ) const
{
    for( size_t a = 0; a < b1.GetCount(); a++ )
    {
        BubbleSim &ea = *b1.GetAt( a );
        for( size_t i = 0; i < b2.GetCount(); i++ )
        {
            BubbleSim &ei = *b2.GetAt( i );
            ( this->*proc ) ( ea, ei );
        }
    }
}

bool CSoapBubblesSim::CollectCellBubbles( tTmpBuffer& buff, CPoint cell )
{
	POSITION pos;
	buff.SetCount( m_Bubbles.GetCellBubbles( cell, pos ) );
	for( size_t i=0; i < buff.GetCount(); i++ )
	{
		BubbleSim &e = m_Bubbles.GetValueAndNextPos( pos );
		buff.SetAt( i, &e );
	}
	return 0 != buff.GetCount();
}

void CSoapBubblesSim::ProcessInteractions( typeBubbleProc proc )
{
    // Proximity-based iterations.
    const CPoint ptMin = m_Bubbles.GetMinCell(), ptMax = m_Bubbles.GetMaxCell();

    // The 2 loops below are by cells.
    for( CPoint pt = ptMin; pt.y <= ptMax.y; pt.y++ )
    {
        for( pt.x = ptMin.x; pt.x <= ptMax.x; pt.x++ )
        {
            if( !CollectCellBubbles( m_tmpThis, pt ) )
                continue;	// No bubbles in this cell at all

            static const CPoint arrNeighbours[ 4 ] = { CPoint( -1, 1 ), CPoint( 1, 1 ), CPoint( 0, 1 ), CPoint( 1, 0 ) };
            for( size_t i=0; i < sizeof( arrNeighbours ) / sizeof( arrNeighbours[0] ); i++ )
            {
                const CPoint ptThat = pt + arrNeighbours[i];
                if( CollectCellBubbles( m_tmpThat, ptThat ) )
                    ProcessInteractionsAcrossBuffers( proc, m_tmpThis, m_tmpThat );
                m_tmpThat.RemoveAll();
            }

            ProcessInteractionsWithinBuffer( proc, m_tmpThis );
        }
    }
}

void CSoapBubblesSim::SimulationStep( float fSecondsPassed )
{
    fSecondsPassed = std::clamp( fSecondsPassed, s_MinFrameTime, s_MaxFrameTime );

	POSITION pos;
	for( m_Bubbles.GetAllBubbles( pos ); NULL != pos; )
	{
		BubbleSim &e = m_Bubbles.GetValueAndNextPos( pos );
		e.curForce = vecType();
        e.nNeighbours = 0;
	}

    ProcessInteractions( &CSoapBubblesSim::BubbleProc_Simulation );

	// Easy part, just convert e.curForce into e.pos
    m_MaxNeighbours = 0;
	for( m_Bubbles.GetAllBubbles( pos ); NULL != pos; )
	{
		const POSITION posCurr = pos;
		BubbleSim &e = m_Bubbles.GetValueAndNextPos( pos );
        e.ConvertForceToMovement( fSecondsPassed );
        m_MaxNeighbours = std::max( m_MaxNeighbours, e.nNeighbours );
        HRESULT hr = m_Bubbles.Update( posCurr );
        ASSERT( SUCCEEDED(hr), "Update() failed.\n%S", LPCTSTR( ErrorMessage(hr) ) );
	}
}

void CSoapBubblesSim::RemoveAround( Vector2D pt, double radius )
{
	BubbleSim bs;
	bs.pos = pt;
	const CPoint cell = bs.GetCell();

	const double distSqr = radius * radius;

	static const CPoint arrNeighbours[9] = {
		CPoint( -1, -1 ), CPoint( 0, -1 ), CPoint( 1, -1 ),
		CPoint( -1, 0 ), CPoint( 0, 0 ), CPoint( 1, 0 ),
		CPoint( -1, 1 ), CPoint( 0, 1 ), CPoint( 1, 1 )
	};

	for( size_t i=0; i < 9; i++ )
	{
		POSITION pos;
		size_t n = m_Bubbles.GetCellBubbles( cell + arrNeighbours[i],  pos );
		if( 0 == n ) continue;
		for( size_t j=0; j<n; j++ )
		{
			POSITION posThis = pos;
			BubbleSim& bThat = m_Bubbles.GetValueAndNextPos(pos);
			if( (bThat.pos - pt).LengthSq > distSqr )
				continue;
			m_Bubbles.Remove( posThis );
		}
	}
}

CRect CSoapBubblesSim::GetBoundingBox()
{
    if( GetBubblesCount() <= 0 )
        return CRect( 0, 0, 0, 0 );

    const CPoint cellMin = m_Bubbles.GetMinCell(), cellMax = m_Bubbles.GetMaxCell() + CPoint( 1, 1 );
    const double left = floor( s_CellSize * cellMin.x );
    const double top = floor( s_CellSize * cellMin.y );
    const double right = ceil( s_CellSize * cellMax.x );
    const double bottom = ceil( s_CellSize * cellMax.y );

    return CRect( int(left), int(top), int(right), int(bottom) );
}

// The main simulation code is over here!

const CSoapBubblesSim::BaseType CSoapBubblesSim::s_CellSize = 45.0;
const CSoapBubblesSim::BaseType CSoapBubblesSim::s_CellSize2 = CSoapBubblesSim::s_CellSize * CSoapBubblesSim::s_CellSize;

const float CSoapBubblesSim::s_MaxFrameTime = 0.05f;
const float CSoapBubblesSim::s_MinFrameTime = 0.001f;

inline CSoapBubblesSim::vecType CSoapBubblesSim::BubbleSim::GetForce( const BubbleSim& that, BaseType& dist ) const
{
    const vecType vDiff = that.pos - pos;
    // http://en.wikipedia.org/wiki/Lennard-Jones_potential says:
    // V(r) = 4 * e * ( (r / s)^-12 - (r/s)^-6 ), so 
    // F(r) = 4 * e * ( -12 * s^12 * r(-13) + 6 * s^6 * r(-7) ) = 
    //   24 * e * ( s^6 * r(-7) - 2 * s^12 * r(-13) );

    const BaseType r2 = vDiff.LengthSq;
    if( r2 >= s_CellSize2 )
    {
        dist = std::numeric_limits<BaseType>::infinity();
        return vecType();
    }

#if 0
    // This branch calculates the force each time

    const BaseType r = sqrt(r2);    // distance between the particles
    const BaseType s = BaseType( size + that.size - 1 ) * BaseType(0.5);    // distance at which the inter-particle potential is zero
    const BaseType e = 1.0;                             // the depth of the potential well
    const BaseType s6 = pow( s, 6 );                    // Saving one pow() at the cost of one multiplication
    const BaseType forceValue = 24.0 * e * ( s6 * pow( r, -7 ) - 2.0 * s6 * s6 * pow( r, -13 ) );

	static const BaseType s_NoiseAmp = BaseType(1E6);
#else
    // This branch calculates the force only once creating the precalculated table.

    static const size_t s_PrecalcElts = 256;
    static BaseType s_dblPrecalc[s_PrecalcElts];
    static bool s_bFirst = true;
	static BaseType s_NoiseAmp = 1.0f;

    if( s_bFirst )
    {
        s_bFirst = false;
		
        for( size_t i=0; i<s_PrecalcElts; i++ )
        {
            const BaseType r = double( 0.5 + i ) * s_CellSize / double( s_PrecalcElts );

            // Achtung!
            // If there're bubbles of different sizes,
            // then we need a separate table for every value of { size + that.size } parameter.

            const BaseType s = BaseType( size + that.size - 1 ) * BaseType(0.5);    // distance at which the inter-particle potential is zero
            const BaseType e = 1.0;                             // the depth of the potential well
            const BaseType s6 = pow( s, 6 );                    // Saving one pow() at the cost of one multiplication
            const BaseType forceValue = 24.0 * e * ( s6 * pow( r, -7 ) - 2.0 * s6 * s6 * pow( r, -13 ) );

            s_dblPrecalc[i] = forceValue;
        }
		s_NoiseAmp = s_dblPrecalc[0];
		s_dblPrecalc[0] = std::numeric_limits<BaseType>::infinity();
    }

    const BaseType r = sqrt( r2 );
    const size_t ind = size_t( ( floor( r ) * double( s_PrecalcElts ) / s_CellSize ) );
    const BaseType forceValue = s_dblPrecalc[ind];
#endif
	dist = r;

	if( _finite(forceValue) )	//< This will work ineffectively if the BaseType will be set to float..
		return vDiff * forceValue;

	// ASSERT( 0, "forceValue is NaN" );
	vecType res;
	do 
	{
		res.x = float( rand() % 0x201 - 0x100 ) / float(0x100);
		res.y = float( rand() % 0x201 - 0x100 ) / float(0x100);
	}
	while( res.LengthSq > 1.0 );

	return res * s_NoiseAmp;
}

void CSoapBubblesSim::BubbleProc_Simulation( BubbleSim& ea, BubbleSim& ei ) const
{
    BaseType dblDistance;
    const vecType force = ea.GetForce( ei, dblDistance );
    CHECK( !_isnan( force.x ) && !_isnan( force.y), return, "Arithmetic failed" );
    ea.curForce += force;
    ei.curForce -= force;

    const double dblThreshold = double( ea.size + ei.size ) * 0.5;

    if( dblDistance > dblThreshold + 2.0 )
        return;
    ei.nNeighbours++;
    ea.nNeighbours++;
}

inline void CSoapBubblesSim::BubbleSim::ConvertForceToMovement( float fSecondsPassed ) 
{
    // This constant represents viscosity ^ (-1)
    const double dblMagicNumber = 0.4;

    vecType vOffset = curForce * ( dblMagicNumber * fSecondsPassed );
    ASSERT( !_isnan(vOffset.x) && !_isnan(vOffset.y), "Arithymetic failed" );

    // This threshold represents the "max pixels per frame" value.
    static const BaseType cMaxOffset = 1.5;

    BaseType LengthSq = vOffset.LengthSq;

    // Modify position, and calculate the floating average energy across frames
    
	if( LengthSq > ( cMaxOffset * cMaxOffset ) )
    {
        vOffset.Length = cMaxOffset;
        LengthSq = cMaxOffset * cMaxOffset;
    }

    pos += vOffset;

	energyCurrentLevel = BYTE( 255.999 * LengthSq / ( cMaxOffset * cMaxOffset ) );

	/*

	// Function Fade( int iOldVal, iCurrLevel ) : iOldVal += valEnergy[ iCurrLevel ]
	static const ptrdiff_t valEnergy[ 9 ] = { -1, 0, 1, 2, 3, 4, 5, 6, 6 };

	// We pretend energyCurrentLevel byte is 3.5 bits fixed-point value,
	// and interpolate between the valEnergy[] control points linearly.

	// Integer part of the value
	const size_t indEnergyInt = energyCurrentLevel >> 5;

	// Fraction part of the value, scaled from 0 to 255
	const size_t indEnergyFrac = ( energyCurrentLevel & 31 ) << 3;

	const ptrdiff_t diff = 
		valEnergy [ indEnergyInt ] << 8 + 
		( valEnergy [ indEnergyInt + 1 ]  - valEnergy [ indEnergyInt ] ) * ( energyCurrentLevel & 31 ) << 3;

	energyFloatingAverage = BYTE( Clamp<ptrdiff_t>( energyFloatingAverage + diff, 0, 0x30000 ) );

	*/

	// Fade factor, 0-256
	const size_t nFadeFactor = 150;

	energyFloatingAverage = ( energyFloatingAverage * ( 256 -  nFadeFactor ) + size_t( energyCurrentLevel ) * nFadeFactor ) >> 8;
}