#pragma once
#include "../Math/Vector2.hpp"
#include "../Math/Vector3.hpp"
#include "../direct3d9/VertexBuffer9.hpp"

#include "BubblesContainer.hpp"
#include "IBubbleSettings.h"

// This class implements the simulation part of the soap bubbles functionality.
class CSoapBubblesSim
{
public:
    typedef double BaseType;
    typedef Vector2<BaseType> vecType;

protected:
	static const BaseType s_CellSize, s_CellSize2;

    static const float s_MinFrameTime, s_MaxFrameTime;

    // This class represents a single bubble, as appears in the physic model.
    struct BubbleSim
    {
		inline BubbleSim()
		{
			energyFloatingAverage = 127 << 8;
		}
        // === Data that persists through the simulation steps === 

        // Position in pixels
        vecType pos;

        // Size (=diameter) in pixels, when the scale is set 1.0
        int size;

        DWORD selfColor;

        BYTE energyCurrentLevel;

        // Floating average bubble energy, from 0 to 255
		DWORD energyFloatingAverage;

        // Count of neighbors; usually it's from 0 to 6.
        BYTE nNeighbours;

        // Get the force between 2 bubbles
		inline vecType GetForce( const BubbleSim& that, BaseType& dist ) const;

        inline void ConvertForceToMovement( float fSecondsPassed );

        // === The following data is only valid during the simulation step, 
        // and is stored here to save both memory allocations and cache misses.
        vecType curForce;

		inline CPoint GetCell() const
		{
			const CPoint cell( static_cast<LONG>( floor( pos.x / s_CellSize ) ), static_cast<LONG>( floor( pos.y / s_CellSize) ) );
			return cell;
		}
    };

	CBubbleContainer<BubbleSim> m_Bubbles;
    BYTE m_MaxNeighbours;

public:
	inline size_t GetBubblesCount() const
	{
		POSITION pos;
		return m_Bubbles.GetAllBubbles( pos );
	}

	void SimulationStep( float fSecondsPassed );

	// Remove all bubbles around the specified point
    void RemoveAround( Vector2D pos, double radius );

    // Get bounding box of all bubbles.
    CRect GetBoundingBox();
private:
	// Same as CAtlArray<BubbleSim*>, but does not free memory when the size is set to 0.
    // This is done to ensure almost no memory [re]allocations will happen in the runtime.
	class tTmpBuffer
	{
		CAtlArray<BubbleSim*> m_arr;
		size_t m_Count;
	public:
		inline tTmpBuffer(){ m_Count = 0; }

        inline void RemoveAll(){ SetCount(0); }

		inline BOOL SetCount( size_t nNewSize )
		{
			m_Count = nNewSize;
			if( 0 == nNewSize  )
				return TRUE;
			return m_arr.SetCount( m_Count );
		}
		inline size_t GetCount(){ return m_Count; }

		inline BubbleSim* GetAt ( size_t i ) { return m_arr[i]; }
		inline void SetAt( size_t i, BubbleSim *e ){ m_arr[i] = e; }
	};

	tTmpBuffer m_tmpThis, m_tmpThat;

	// Returns false if no bubbles are in that cell,
	// otherwise collects the bubbles from that cell, storing them in the tTmpBuffer
	bool CollectCellBubbles( tTmpBuffer& buff, CPoint cell );

    // This 'typeBubbleProc' is for 'SimulationStep' functionality.
    void BubbleProc_Simulation( BubbleSim& ea, BubbleSim& ei ) const;

protected:
    // The stuff below implements the proximity-based bubbles iterations.

    // NB!
    // If it was already called for { a, b } pair, it's <b>not</b> going to be called for (b, a) pair.
    typedef void ( CSoapBubblesSim::*typeBubbleProc )( BubbleSim& e1, BubbleSim& e2 ) const;

    // Call 'proc' exactly once for every bubbles unordered pair { i, j }
    // so that i &isin; buff, j &isin; buff, and i &ne; j.
    void ProcessInteractionsWithinBuffer( typeBubbleProc proc, tTmpBuffer& buff ) const;

    // Call 'proc' exactly once for every bubbles unordered pair { i, j },
    // so that i &isin; buff1, j &isin; buff2.
    void ProcessInteractionsAcrossBuffers( typeBubbleProc proc, tTmpBuffer& buff1, tTmpBuffer& buff2 ) const;

    // Call 'proc' exactly once for every bubbles unordered pair { i, j } so that i &ne; j.
    // Only interactions within CBubbleContainer's cell, and between adjacent cells, are being processed.
    // Both 4-adjacency and 8-adjacency is being processed.
    void ProcessInteractions( typeBubbleProc proc );

private:
    typedef CAtlMap< BubbleSim*, size_t > tTmpMap;
    mutable tTmpMap m_tmpNeighboursMap;
    void BubbleProc_PaintStructure( BubbleSim& ea, BubbleSim& ei ) const;
};

// This class implements the rendering part of the soap bubbles functionality
class CSoapBubblesRender:
    public CSoapBubblesSim
{
public:
    CSoapBubblesRender();
    ~CSoapBubblesRender();

    HRESULT BufferCreate( IDirect3DDevice9* pDevice, D3DPOOL Pool = D3DPOOL_DEFAULT );
    HRESULT BufferDestroy();

    // Update the stored buffer, and call SetStreamSource
    // The sufficient vertex buffer must be already allocated with BufferCreate() method
    HRESULT BufferUpdate();

    HRESULT BufferSetStreamSource( UINT nStream = 0 );

private:
    struct BubbleRender
    {
        Vector3F pos;       // untransformed position
        float pxSize;       // vertex point size
        DWORD diffuseColor; // diffuse color in ARGB

        static DWORD GetFVF() { return D3DFVF_XYZ | D3DFVF_PSIZE | D3DFVF_DIFFUSE; }
    };

    CVertexBuffer<BubbleRender> m_vb;
    IDirect3DDevice9* m_pDevice;
    D3DPOOL m_pool;

    BubbleRender PrepareBubbleForRender( const CSoapBubblesSim::BubbleSim& bs );
protected:
    ColoringParams m_BubblesColoring;
};

// This class implements the bubbles operations.
class CSoapBubblesModel:
    public CSoapBubblesRender,
    public IBubbleSettings
{
public:
    static const int s_Size = 10;

    inline CSoapBubblesModel(){}
    inline ~CSoapBubblesModel(){}

    // 
    HRESULT OnMouse( UINT vk, CPoint pt )
    {
        return E_NOTIMPL;
    }

    // Add 'nBubbles' bubbles, randomly-colored, randomly distributed within the ( -sz/2; +sz/2 ) rectangle
    HRESULT AddRandomBubbles( size_t nBubbles, CSize sz );

    HRESULT MouseOperation( UINT vk, Vector2D worldPos );

private:
    // IBubbleSettings implementation.
    virtual HRESULT GetColoring( ColoringParams* params )
    {
        *params = m_BubblesColoring;
        return S_OK;
    }

    virtual HRESULT SetColoring( const ColoringParams* params )
    {
        m_BubblesColoring = *params;
        return S_OK;
    }

    virtual HRESULT GetClickActions( ClickActions* params )
    {
        return E_NOTIMPL;
    }

    virtual HRESULT SetClickActions( const ClickActions* params )
    {
        return E_NOTIMPL;
    }
};