#pragma once
#include "../VsTabs/TabViewImpl.hpp"
#include "GdiResources.h"

// This view displays 32BPP RGBA image, and supports zooming and scrolling.
// Instances of this class holds 32BPP pre-multiplied alpha image in RAM.
// So don't use it for 50 MPix images :-)
class CImageWnd:
	public NVsTabs::CTabViewWindowImpl<CImageWnd>
{
public:
    CImageWnd( NVsTabs::CTabsOwner* pOwner );
    virtual ~CImageWnd();

	HRESULT Create( CWindow* pParent );

    // Create an image of the specified size, allocating the necessary buffer/bitmap
    // Safe to be called from different threads
    HRESULT InitImage(CSize sz);

	static const DDARGB s_EmptyColor;

    // Set the bitmap data for the specified line of pixels,
    // and invalidate the corresponding portion of the window.
    HRESULT SetBitmapLine(int iLine, const DDARGB* pPixels);

    // Size of the bitmap in pixels
    inline CSize GetSize() const { return m_memSize; }
    __declspec(property(get=GetSize)) CSize Size;
public:
    static const DWORD dwClassStyle = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
    static ATL::CWndClassInfo& GetWndClassInfo()
    {
        static ATL::CWndClassInfo wc =
        {
            { sizeof(WNDCLASSEX), dwClassStyle, StartWindowProc, 0, 0, NULL, NULL, NULL,
            (HBRUSH)((size_t)GetStockObject(DKGRAY_BRUSH) + 1), NULL, L"ImageWnd", NULL }, 
            NULL, NULL, NULL, TRUE, 0, _T("") 
        };
        return wc;
    }

    BEGIN_MSG_MAP( CImageWnd )
		CHAIN_MSG_MAP( TTabViewImpl )
    // General
        MESSAGE_HANDLER( WM_SIZE,          OnSize );
        MESSAGE_HANDLER( WM_PAINT,         OnPaint );
    // Zooming
        MESSAGE_HANDLER( WM_CHAR,          OnChar );
        MESSAGE_HANDLER( WM_MBUTTONUP,     OnMButtonUp );
        MESSAGE_HANDLER( WM_MOUSEWHEEL,    OnMouseWheel );

    // Scrolling
        MESSAGE_HANDLER( WM_MOUSEMOVE,     OnWmMouseMove )
        MESSAGE_HANDLER( WM_MOUSELEAVE,    OnWmMouseLeave )
        MESSAGE_HANDLER( WM_LBUTTONDOWN,   OnWmMouseDown )
        MESSAGE_HANDLER( WM_LBUTTONUP,     OnWmMouseUp )
    END_MSG_MAP()
private:
    // Called from InitImage().
    void FillNewImage();

    LRESULT OnSize( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );
    LRESULT OnPaint( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );

    LRESULT OnChar( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );
    LRESULT OnMouseWheel( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );
    LRESULT OnMButtonUp( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );

    LRESULT OnWmMouseMove( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );
    LRESULT OnWmMouseLeave( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );
    LRESULT OnWmMouseDown( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );
    LRESULT OnWmMouseUp( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );

    bool m_bMouseTracking, m_bMouseDrag;
    CPoint m_ptMouseDown;
    
    // Available zoom levels
    struct ZoomLevel
    {
        int iMul;
        int iDiv;
    };
    static const ZoomLevel s_ZoomLevels[];
    static const int s_DefaultZoomLevel;

    // Size of an image, in pixels, before any zooming (as stored in memory)
    CSize m_memSize;

    // ==== Stuff for painting ====
    // The memory bitmap is pre-multiplied RGBA, size is exactly m_Size.
    HBITMAP m_hMemBitmap;
    RGBQUAD *m_pMemBitmapData;
    HDC m_hMemDC;
    HBITMAP m_hPrevBitmap;

    // ==== Current display position and size ====
    // Zoom level ( index in s_ZoomLevels )
    int m_ZoomLevel;
    // Top-left point of an image in client coordinates
    CPoint m_clientTopTeft;
    // TODO: scroll bars are not implemented yet.
    bool m_bHScroll, m_bVScroll;

    CGdiResources m_res;
    CRect m_ClientRect;

    // Wrapper around MSIMG32.DLL::AlphaBlend
    BOOL AlphaBlend( HDC dcDest, CRect& rcDest, HDC dcSrc, CRect rcSrc );

    // Convert size between client and memory bitmap
    CSize memFromClientSz( CSize clientSz );
    CSize clientFromMemSz( CSize memSz );

    // Convert position between client and memory bitmap
    CPoint memFromClientPt( CPoint clientPt );
    CPoint clientFromMemPt( CPoint memPt );
    CRect clientFromMemRect( CRect rc );

    // The ::IntersectRect fails when source==destination,
    // so CRect::IntersectRect fails when this==Rect1 or this==Rect2 :-(
    inline CRect IntersectRect( const CRect& r1, const CRect& r2 )
    {
        CRect res;
        res.IntersectRect( r1, r2 );
        return res;
    }

    CRect PlaceRectInsideOrCenter( const CRect& RectToPlace, const CRect& OuterArea );

    // Set the current zoom level.
    // The clientAhchor is the point in client coordinates,
    // which must remain on its place as the scrolling occurs
    void SetZoomLevel( int iNewZoomLevel, CPoint clientAhchor );

    inline RGBQUAD PremultiplyAlpha( DDARGB c )
    {
        RGBQUAD res;
        res.rgbReserved = c.alpha;
        res.rgbRed   = BYTE( int( c.red ) * int( c.alpha ) / 255 );
        res.rgbGreen = BYTE( int( c.green ) * int( c.alpha ) / 255 );
        res.rgbBlue  = BYTE( int( c.blue ) * int( c.alpha ) / 255 );
        return res;
    }

    // Expand the dirty rectangle to fit the pixel grid in the memory bitmap.
    // It only changes the rectangle if zoom > 100%
    void ExpandDirtyRect( CRect& rc );
};