#include "stdafx.h"
#include "UnitTest.h"

#if UNIT_TESTS

class CUnitTestsImpl
{
	struct TEST_IMPL
	{
		CStringA strName;
		UnitTestProc proc;
	};
	typedef CAtlArray<TEST_IMPL> ArrayType;
	typedef CAtlMap<CStringA, ArrayType> MapType;
	MapType m_tests;

	enum E_TEST_RES
	{
		TEST_OK = 0,
		TEST_FAILED_HR = 1,
		TEST_CPP_EXCEPTION = 2,
		TEST_SEH = 3
	};

	E_TEST_RES RunTest( UnitTestProc proc, HRESULT& hr, bool bSilent );
	E_TEST_RES RunTest2( UnitTestProc proc, HRESULT& hr );

	int RunAllImpl( FILE* f, bool bSilent, int* pTestsCount );

	CUnitTestsImpl(){}
	~CUnitTestsImpl(){}
public:
	static CUnitTestsImpl &This()
	{
		static CUnitTestsImpl res;
		return res;
	}

	void Register( const char* strFile, const char* strName, UnitTestProc testProc )
	{
		const TEST_IMPL ti = {CStringA(strName), testProc };
		m_tests[strFile].Add(ti);
	}

	static int RunAll( FILE* f, bool bSilent, int* pTestsCount )
    {
        return This().RunAllImpl( f, bSilent, pTestsCount );
    }
};

void RegisterUnitTest( const char* strFile, const char* strName, UnitTestProc testProc )
{
	CUnitTestsImpl::This().Register( strFile, strName, testProc );
}

CUnitTestsImpl::E_TEST_RES CUnitTestsImpl::RunTest( UnitTestProc proc, HRESULT& hr, bool bSilent )
{
	if( !bSilent )
	{
		hr = proc();
		return SUCCEEDED(hr) ? TEST_OK : TEST_FAILED_HR;
	}

	__try
	{
		return RunTest2( proc, hr );
	}
	__except( EXCEPTION_EXECUTE_HANDLER )
	{
		return TEST_SEH;
	}
}

CUnitTestsImpl::E_TEST_RES CUnitTestsImpl::RunTest2( UnitTestProc proc, HRESULT& hr )
{
	hr = E_UNEXPECTED;
	try
	{
		return ( SUCCEEDED( hr = proc() ) ) ? TEST_OK : TEST_FAILED_HR;
	}
	catch(...)
	{
		return TEST_CPP_EXCEPTION;
	}
}

int CUnitTestsImpl::RunAllImpl( FILE* stream, bool bSilent, int* pTestsCount )
{
	int iTotal = 0;
	int iFailed = 0;
	for( POSITION p = m_tests.GetStartPosition(); p; )
	{
		const MapType::CPair* f = m_tests.GetNext(p);
		fprintf( stream, "\tFile: %s\n", (const char*)( f->m_key) );
		const ArrayType& a = f->m_value;
		for( size_t i = 0; i < a.GetCount(); i++ )
		{
			iTotal++;
			HRESULT hr = S_OK;
			const E_TEST_RES res = RunTest( a[i].proc, hr, bSilent );
			fprintf( stream, "%i\t%s\t", int(res), (const char*)( a[i].strName ) );

			if( TEST_OK == res )
			{
				fprintf( stream, "OK\t0x%.8X\n", int( hr ) );
				continue;
			}
			iFailed++;
			switch( res )
			{
			case TEST_FAILED_HR:
				fprintf( stream, "FAILED\t0x%.8X, %S\n", int( hr ), LPCTSTR( ErrorMessage( hr ) ) );
				break;
			case TEST_CPP_EXCEPTION:
				fprintf( stream, "FAILED\tC++ exception\n" );
				break;
			case TEST_SEH:
				fprintf( stream, "FAILED\tstructured exception\n" );
				break;
			default:
				__debugbreak();
			}
		}
	}
	if(pTestsCount)
		*pTestsCount = iTotal;
	return iFailed;
}

int RunUnitTests( bool bSilent )
{
	// Open the output file
	CStringA strFileName;
	if( !GetModuleFileNameA( NULL, strFileName.GetBufferSetLength( MAX_PATH ), MAX_PATH ) )
		return -1;
	strFileName.ReleaseBuffer();
	strFileName += ".UnitTests.txt";

	FILE* f = fopen( strFileName, "w" );
	if( NULL == f )
		return -2;

	// Write the date-time string
	CStringA strDateTime;
	const CTime t = CTime::GetCurrentTime();
	const CString strTimeString = t.Format( L"%Y %B, %d (%A), %H:%M:%S" );
	fprintf( f, "\t\t%S\n", LPCTSTR(strTimeString) );
	fflush( f );

	// Run the tests
	int iTests;
	const int res = CUnitTestsImpl::RunAll( f, bSilent, &iTests );
	fclose( f );

	if( !bSilent )
	{
		if( !res )
		{
			CString msg;
			msg.Format( L"All %i tests passed OK.\nEnjoy.\n\nThe following text file contains the tests log:\n%S",
                iTests, (const char*)(strFileName) );
			::MessageBox(NULL, msg, L"Unit Tests OK", MB_OK | MB_ICONINFORMATION );
		}
		else
		{
			CString msg;
			msg.Format( L"%i of %i unit tests failed.\nThe following text file contains the tests log:\n%S", res, iTests, (const char*)(strFileName) );
			::MessageBox(NULL, msg, L"Unit Tests Failed", MB_OK | MB_ICONWARNING );
		}
	}
	return res;
}
#endif
