#include "StdAfx.h"
#include "VSTabs.h"

CVSTabs::CVSTabs(void):
	m_btnLeft(DFC_SCROLL, DFCS_SCROLLLEFT),
	m_btnRight(DFC_SCROLL, DFCS_SCROLLRIGHT),
	m_btnClose(DFC_CAPTION, DFCS_CAPTIONCLOSE)
{
	m_pButtons[0] = &m_btnClose;
	m_pButtons[1] = &m_btnRight;
	m_pButtons[2] = &m_btnLeft;
	m_FirstTab = m_LastTab = 0;
    m_ActiveTab = -1;

	m_fontInactive = m_fontActive = NULL;
	m_pnOutline = NULL;

	m_dwNextTabID = 1;
}

#pragma region CVSTabs_TabInfo

CVSTabs::TabInfo::TabInfo()
{
	Region = NULL;
	dwID = 0;
}

CVSTabs::TabInfo::~TabInfo()
{
	if( NULL != Region )
	{
		DeleteObject(Region);
		Region = NULL;
	}
}

void CVSTabs::TabInfo::CopyFrom( const TabInfo& that )
{
	Caption = that.Caption;
	pxCaptionLength = that.pxCaptionLength;
	Tooltip = that.Tooltip;
	CopyMemory( &Border, that.Border, sizeof( Border ) );
	if( NULL != Region )
	{
		// Will be recreated during the next RecalcLayout
		DeleteObject(Region);
		Region = NULL;
	}
	dwID = that.dwID;
    pUserData = that.pUserData;
}

bool CVSTabs::TabInfo::SetText(HDC dc, LPCTSTR strNewText)
{
	CSize szText;
	Caption = strNewText;
	if(!GetTextExtentPoint32(dc, Caption, Caption.GetLength(), &szText))
		return false;
	pxCaptionLength = szText.cx;
	return true;
}

bool CVSTabs::TabInfo::RecalcLayout(int cy, int *xLeftMost, int* xEnd)
{
	const int iTop = 1;	                   // Top edge
	const int iBottom = cy-1;              // Bottom edge
	const int iOvr = cy/2 + iBottom - cy;  // Overlapping
	const int iRound = 3;                  // Rounding borders

    pxCaptionLength = std::max(pxCaptionLength, 4 * iRound);
	Border[0].SetPoint(*xLeftMost, iBottom);            // Left point at the bottom
	Border[1] = Border[0] + CPoint(iOvr, -iOvr);        // Intersection with the next overlapping one
	const int i2 = iBottom - iTop - iRound;
	Border[2] = Border[0] + CPoint(i2, -i2);            // Start of the left chamfer
	Border[3] = Border[2] + CPoint(2*iRound, -iRound);  // Left point on the top
	Border[4].x = Border[2].x + pxCaptionLength + 2;
	Border[4].y = Border[3].y;                          // Right point on the top
	Border[5] = Border[4] + CPoint(iRound, iRound);     // End of the right chamfer
	Border[6].x = Border[5].x;
	Border[6].y = iBottom;                              // Right point at the bottom

	if(Region)
	{
		DeleteObject(Region);
		Region = NULL;
	}

	*xLeftMost = Border[6].x - iOvr;
	*xEnd = Border[6].x;

	return true;
}

int CVSTabs::TabInfo::GetWidth(bool bIncludeOverflow) const
{
	if(bIncludeOverflow)
		return Border[6].x - Border[0].x;
	return Border[6].x - Border[1].x;
}

CRect CVSTabs::TabInfo::GetRect(bool bIncludeOverflow) const
{
	CRect res;
	res.left = ( bIncludeOverflow ) ? Border[0].x : Border[1].x;
	res.right = Border[6].x;
	res.top = Border[3].y;
	res.bottom = Border[0].y;
	return res;
}

int CVSTabs::TabInfo::GetRight() const
{
	return Border[6].x;
}

bool CVSTabs::TabInfo::RecalcRegion() const
{
	if( NULL != Region )
	{
		::DeleteObject( Region );
		Region = NULL;
	}
		
	if( !( Region = CreatePolygonRgn( Border, 7, WINDING ) ) )
		return false;
	return true;
}

bool CVSTabs::TabInfo::FillTabRegion( HDC dc, HBRUSH br ) const
{
	if( NULL == Region )
		if( !RecalcRegion() )
			return false;

	if( !FillRgn( dc, Region, br ) )
		return false;
	return true;
}

int CVSTabs::TabInfo::GetTextCenter() const
{
	return ( Border[2].x + Border[5].x ) / 2;
}

bool CVSTabs::TabInfo::DrawFrameAndText( HDC dc, bool bLeftSideOverlapped ) const
{
	int nPoints;
	CPoint ptFirst;
	const CPoint *pLineFirst;
	if( bLeftSideOverlapped )
	{
		ptFirst = Border[1];
		nPoints = 5;
		pLineFirst = Border + 2;
	}
	else
	{
		ptFirst = Border[0];
		ptFirst.x++;
		ptFirst.y--;
		nPoints = 6;
		pLineFirst = Border + 1;
	}

	if( !MoveToEx( dc, ptFirst.x, ptFirst.y, NULL) )
		return false;
	if( !PolylineTo( dc, pLineFirst, nPoints) )
		return false;

	const int x = ( Border[2].x + Border[5].x ) / 2 + 1;
	const int y = Border[0].y - 2;
	if( !TextOut( dc, x, y, Caption, Caption.GetLength() ) )
		return false;
	return true;
}

bool CVSTabs::TabInfo::HitTest(CPoint client) const
{
	CRect rc = GetRect( true );
	if( !rc.PtInRect( client ) )
		return false;
	if ( !PtInRegion( Region, client.x, client.y ) )
		return false;
	return true;
}
#pragma endregion

CVSTabs::~CVSTabs(void)
{
}
HRESULT CVSTabs::Create(CWindow* pParent, IVSTabsHost* pOwner, UINT uID)
{
	if( NULL == pParent || NULL == pOwner )
		return E_POINTER;
	m_pOwner = pOwner;
	HWND w = __super::Create(pParent->m_hWnd, NULL, L"VSTabs", WS_CHILD | WS_VISIBLE, 0, uID);
	if(!w)
		return GetLastHr();
	ShowWindow(SW_SHOW);
	return S_OK;
}

LRESULT CVSTabs::OnCreate(UINT nMsg, WPARAM wParam,  LPARAM lParam, BOOL& bHandled)
{
	if(!RecalcMetrics())
	{
		ATLTRACE("CVSTabs::OnCreate: RecalcMetrics() failed.");
		return -1;
	}

	if(!(m_pnOutline = CreatePen(PS_SOLID, 0, GetSysColor(COLOR_BTNSHADOW))))
	{
		ATLTRACE("CVSTabs::OnCreate: unable to create COLOR_BTNSHADOW pen");
		return -1;
	}
	m_brActiveTab = GetSysColorBrush(COLOR_WINDOW);
	m_colText = GetSysColor(COLOR_BTNTEXT);

	return 0;
}

LRESULT CVSTabs::OnSize(UINT nMsg, WPARAM wParam,  LPARAM lParam, BOOL& bHandled)
{
	m_bPositionsDirty = true;
	InvalidateRect(0, TRUE);
	return 0;
}

LRESULT CVSTabs::OnDestroy(UINT nMsg, WPARAM wParam,  LPARAM lParam, BOOL& bHandled)
{
	if(m_fontInactive)
		DeleteObject(m_fontInactive);
	if(m_fontActive)
		DeleteObject(m_fontActive);

	DeleteObject(m_pnOutline);
	return 0;
}

BOOL CVSTabs::RecalcMetrics()
{
	NONCLIENTMETRICS metrics;
	const UINT cbSize = sizeof( metrics );
	ZeroMemory( &metrics, cbSize );
	metrics.cbSize = cbSize;
	if( !SystemParametersInfo( SPI_GETNONCLIENTMETRICS, cbSize, &metrics, 0 ) )
	{
		ATLTRACE( "CVSTabs::RecalcMetrics: SystemParametersInfo() failed." );
		return FALSE;
	}

	HFONT fActive, fInactive;
	LOGFONT Font;
	CopyMemory( &Font, &metrics.lfMenuFont, sizeof( Font ) );
	Font.lfWeight = FW_SEMIBOLD;
	fActive = CreateFontIndirect( &Font );
	Font.lfWeight = FW_NORMAL;
	fInactive = CreateFontIndirect( &Font );
	if( !fActive || !fInactive )
	{
		ATLTRACE( "CVSTabs::RecalcMetrics: unable to create a font" );
		return FALSE;
	}

	if( m_fontInactive )
		DeleteObject( m_fontInactive );
	m_fontInactive = fInactive;

	if( m_fontActive )
		DeleteObject( m_fontActive );
	m_fontActive = fActive;

	m_TotalHeight = metrics.iMenuHeight + 1;
	
//	m_szButtons.SetSize( metrics.iCaptionWidth-2, metrics.iCaptionHeight-2 );
//	m_szButtons.SetSize( metrics.iCaptionWidth, metrics.iCaptionHeight );
//	m_TotalHeight = max( metrics.iMenuHeight, m_szButtons.cy );
	m_szButtons.SetSize( m_TotalHeight-4, m_TotalHeight-4 );

	m_bPositionsDirty = true;

	return TRUE;
}

void CVSTabs::ButtonInfo::RecalcLayout( CSize sz, int iRight, CPoint ptOffset )
{
	m_rc = CRect( CPoint( 0, 0 ), sz );
	m_rc.MoveToX( iRight - sz.cx - ptOffset.x );
	m_rc.MoveToY( ptOffset.y );
}

void CVSTabs::ButtonInfo::Draw( HDC dc ) const
{
    UINT uState = m_dfcFlag;
    uState |= ( DFCS_FLAT | DFCS_TRANSPARENT );
    if( !m_bEnabled )
        uState |= DFCS_INACTIVE;
    else if( m_bDown )
        uState |= ( DFCS_PUSHED | DFCS_HOT );
    else if( m_bOver )
        uState |= DFCS_HOT;

    CRect rc = m_rc;
    DrawFrameControl( dc, rc, m_dfcType, uState );
}

void CVSTabs::RecalcLayout()
{
	CRect rc;
	GetClientRect(&rc);

	const int iButtonSpace = ( m_TotalHeight - m_szButtons.cy ) / 2;
	const CPoint btnOffset( iButtonSpace, iButtonSpace );

	m_btnClose.RecalcLayout( m_szButtons, rc.right, btnOffset );
	m_btnClose.m_bVisible = true;
    m_btnClose.m_bEnabled = ( m_tabs.GetCount() > 0 );
	m_btnRight.RecalcLayout( m_szButtons, m_btnClose.m_rc.left, btnOffset );
	m_btnLeft.RecalcLayout( m_szButtons, m_btnRight.m_rc.left, btnOffset );

	m_ActiveTabNeedsShrinkage = false;
	int iMaxRight = m_btnClose.m_rc.left - iButtonSpace;
	const int iTotalTabs = int(m_tabs.GetCount());

	const int iLeftFree = 3;
	// First we pretend every tab fits the width
	int iNext=iLeftFree, iEnd=0;
	for( int i=0; i<iTotalTabs; i++ )
		m_tabs[i].RecalcLayout(m_TotalHeight, &iNext, &iEnd);

	if( iEnd <= iMaxRight )	// All fine! :-)
	{
		m_FirstTab=0;
		m_LastTab=iTotalTabs-1;
		m_btnRight.m_bVisible = m_btnLeft.m_bVisible = false;
	}
	else
	{	// Don't fit.
		iMaxRight = m_btnLeft.m_rc.left - iButtonSpace;

		do 
		{
			int iTotalWidth = m_tabs[m_ActiveTab].GetWidth(true) + iLeftFree;

			// Try active tab only
			if( iTotalWidth > iMaxRight ) // Worse scenario: even the single active tab doesn't fit :-(
			{
				m_ActiveTabNeedsShrinkage = true;
				m_FirstTab = m_LastTab = m_ActiveTab;
				break;
			}

			// move left to the m_FirstTab
			bool bBreak = false;
			int i;
			for( i=m_ActiveTab-1; i>=m_FirstTab; i--)
			{
				iTotalWidth += m_tabs[i].GetWidth(false);
				if( iTotalWidth > iMaxRight )
					break;
			}
			if( iTotalWidth > iMaxRight )
			{
				m_FirstTab = i+1;
				m_LastTab = m_ActiveTab;
				break;
			}
			// OK, now we sure tabs from m_First to m_Active fits.

			// Move right to the iTotalTabs
			for( i=m_ActiveTab+1; i<iTotalTabs; i++ )
			{
				iTotalWidth += m_tabs[i].GetWidth(false);
				if( iTotalWidth > iMaxRight )
					break;
			}
			if( iTotalWidth > iMaxRight )
			{
				m_LastTab = i-1;
				break;
			}
			m_LastTab = iTotalTabs-1;

			// Move left from m_FirstTab to 0
			for( i=m_FirstTab-1; i>=0; i-- )
			{
				iTotalWidth += m_tabs[i].GetWidth(false);
				if( iTotalWidth > iMaxRight )
					break;
			}
            ASSERT( iTotalWidth > iMaxRight, "The pixel size arithmetic failed somehow." );
			m_FirstTab = i+1;
		}
		while(0);

		// Recalculate the layout using first/last tabs found/updated
		iNext=iLeftFree;
		for( int i=m_FirstTab; i<=m_LastTab; i++ )
			m_tabs[i].RecalcLayout( m_TotalHeight, &iNext, &iEnd );
        ASSERT( iEnd <= iMaxRight, "Even the singe tab does not fit. Consider modyfying this class to truncate the active tab in this case." );

		// enable/disable the buttons
		m_btnRight.m_bVisible = m_btnLeft.m_bVisible = true;
		m_btnLeft.m_bEnabled = ( m_FirstTab > 0 );
		m_btnRight.m_bEnabled = ( m_LastTab < iTotalTabs - 1 );
	}

	// recalculate the tabs regions
	for( int i=m_FirstTab; i<=m_LastTab; i++ )
		m_tabs[i].RecalcRegion();

	m_bPositionsDirty = false;
	return;

}

LRESULT CVSTabs::OnPaint(UINT nMsg, WPARAM wParam,  LPARAM lParam, BOOL& bHandled)
{
	PAINTSTRUCT ps;
	HDC dc = BeginPaint(&ps);

	if(m_bPositionsDirty)
		RecalcLayout();

	SelectObject( dc, m_pnOutline );

	SetBkMode( dc, TRANSPARENT ); 
	SetTextColor( dc, m_colText );

	// Thin bottom line
	MoveToEx( dc, ps.rcPaint.left, m_TotalHeight - 2, NULL );
	LineTo( dc, ps.rcPaint.right, m_TotalHeight - 2 );
	
	// The buttons
	const ButtonInfo *pButtons[3] = { &m_btnLeft, &m_btnRight, &m_btnClose };
	for( int i=0; i<3; i++ )
	{
		const ButtonInfo b = *pButtons[i];
		if( !b.m_bVisible )
			continue;
		CRect rc;
		rc.IntersectRect( &ps.rcPaint, b.m_rc );
		if( rc.IsRectEmpty() )
			continue;
        b.Draw( dc );
	}

	if( m_tabs.GetCount() < 1 || m_tabs[m_LastTab].GetRight() < ps.rcPaint.left )
	{
		EndPaint( &ps );
		return 0;
	}

	//TODO: check is m_ActiveTabNeedsShrinkage is set, and paint accordingly
    
    // Update the tabs
	SetTextAlign( dc, TA_CENTER | TA_BOTTOM );
	bool bLeftOverlapped = false;
	for( int i=m_FirstTab; i<=m_LastTab; i++ )
	{
		CRect rc;
		rc.IntersectRect( &ps.rcPaint, m_tabs[i].GetRect( true ) );
		if( !rc.IsRectEmpty() )
		{
			if( i == m_ActiveTab )
			{
				bLeftOverlapped = false;
				m_tabs[i].FillTabRegion( dc, m_brActiveTab );
				SelectObject( dc, m_fontActive );
			}
			else
				SelectObject( dc, m_fontInactive );
			m_tabs[i].DrawFrameAndText( dc, bLeftOverlapped );
		}
		bLeftOverlapped = true;
	}

	EndPaint( &ps );
	return 0;
}

int CVSTabs::HitTest( CPoint pt )
{
	if( pt.x < 0 )
		return -1;
	for( int i=0; i<s_Buttons; i++)
	{
		if( m_pButtons[i]->m_rc.PtInRect( pt ) )
			return -2-i;
	}
	int iResult = -1;
	for( int i=m_FirstTab; i<=m_LastTab; i++ )
	{
		if( !m_tabs[i].HitTest( pt ) )
			continue;
		if( iResult < 0 || i == m_ActiveTab )
			iResult = i;
	}
	return iResult;
}

void CVSTabs::OnMouseInOut( int iCtrl, bool bIn )
{
	if( iCtrl >= -1 ) return;

	const int iButton = -2-iCtrl;
	m_pButtons[iButton]->m_bOver = bIn;
	InvalidateRect( m_pButtons[iButton]->m_rc );
}

void CVSTabs::OnMouseHover( int iCtrl, CPoint pt )
{
// TODO: tooltip
}

void CVSTabs::OnMouseDown( int iCtrl, DWORD vk )
{
	if( iCtrl >= -1 || VK_RBUTTON == vk ) return;

	const int iButton = -2-iCtrl;
	m_pButtons[iButton]->m_bDown = true;
	InvalidateRect( m_pButtons[iButton]->m_rc );
}

void CVSTabs::OnMouseUp( int iCtrl, DWORD vk )
{
	if( iCtrl >= -1 || VK_RBUTTON == vk ) return;

	const int iButton = -2-iCtrl;
	m_pButtons[iButton]->m_bDown = false;
	InvalidateRect( m_pButtons[iButton]->m_rc );
}

void CVSTabs::OnMouseClick( int iCtrl, DWORD vk )
{
	if( iCtrl== -1 )
		return;	// Non-occupied space

	if( iCtrl>=0 && VK_LBUTTON == vk )
	{
		if( m_ActiveTab == iCtrl )
			return;
		InvalidateRect( m_tabs[m_ActiveTab].GetRect( true ) );
		InvalidateRect( m_tabs[iCtrl].GetRect( true ) );
		m_ActiveTab = iCtrl;
		m_pOwner->OnTabSwitched( m_tabs[iCtrl].dwID );
		return;
	}

	if( iCtrl>=0 && VK_RBUTTON == vk )
	{
		// TODO: Context menu
		return;
	}

	if( iCtrl < 0 && VK_LBUTTON == vk )
	{
		const int iButton = -2-iCtrl;
		if( !m_pButtons[iButton]->m_bEnabled )
			return;
		if( 0 == iButton )	// 'Close' button
			m_pOwner->OnTabClose( m_tabs[m_ActiveTab].dwID );
		if ( 1 == iButton)	// 'Right' button
		{
			if( m_LastTab+1 == m_tabs.GetCount() )
				return;
			m_LastTab++;
			m_FirstTab++;
			if( m_ActiveTab == m_FirstTab-1 )
			{
				m_ActiveTab = m_FirstTab;
				m_pOwner->OnTabSwitched( m_tabs[m_ActiveTab].dwID );
			}
			m_bPositionsDirty = true;
			InvalidateRect( NULL );
			return;
		}
		if ( 2 == iButton)	// Left button
		{
			if( m_FirstTab == 0 )
				return;
			m_LastTab--;
			m_FirstTab--;
			if( m_ActiveTab == m_LastTab+1 )
			{
				m_ActiveTab = m_LastTab;
				m_pOwner->OnTabSwitched( m_tabs[m_ActiveTab].dwID );
			}

			m_bPositionsDirty = true;
			InvalidateRect(NULL);
			return;
		}
	}
}

// 
DWORD CVSTabs::AddTab(LPCTSTR strText, void* pUserData)
{
	const DWORD dwNewID = m_dwNextTabID;
	m_dwNextTabID++;
	TabInfo NewTab;
	NewTab.dwID = dwNewID;
    NewTab.pUserData = pUserData;
    const int szNewIndex = int( m_tabs.Add(NewTab) );
    m_tabsLookup[dwNewID] = szNewIndex;
	SetTabText( dwNewID, strText );
	m_bPositionsDirty = true;
    if( 1 == m_tabs.GetCount() )
        SwitchToTab( dwNewID );
    else
        InvalidateRect( NULL, TRUE );
	return dwNewID;
}

void CVSTabs::SetTabText( DWORD dwID, LPCTSTR strText )
{
	TabInfo* pTab = FindTabByID( dwID );
	if(!pTab) return;

	HDC dc = GetDC();
	SelectObject( dc, m_fontActive );
	pTab->SetText( dc, strText );
	ReleaseDC( dc );
	m_bPositionsDirty = true;
	InvalidateRect( NULL );
}

void CVSTabs::SwitchToTab( DWORD dwID )
{
	TabInfo* pTab = FindTabByID( dwID );
	if( !pTab ) return;
	int iTab = int( pTab - m_tabs.GetData() );
	m_ActiveTab = iTab;
	m_bPositionsDirty = true;
	InvalidateRect(NULL);
	m_pOwner->OnTabSwitched(dwID);
}

void CVSTabs::RemoveTab( DWORD dwID )
{
	TabInfo* pTab = FindTabByID( dwID );
	if( !pTab ) return;

    const int iTab = int( pTab - m_tabs.GetData() );
    m_tabsLookup.RemoveKey( dwID );

    // Fix the indices in m_tabsLookup
    for( POSITION pos = m_tabsLookup.GetStartPosition(); pos; )
    {
        MapType::CPair* p = m_tabsLookup.GetNext( pos );
        if( p->m_value > iTab )
            p->m_value--;
    }

    m_tabs.RemoveAt( iTab );

	if( iTab < m_FirstTab )   //< the closed tab was hidden in the left
	{
		m_FirstTab--;
		m_ActiveTab--;
		m_LastTab--;
	}
	else if( iTab <= m_LastTab )    //< the closed tab was visible..
	{
		if( iTab == m_ActiveTab ) // ..and active one
		{
			m_ActiveTab = std::max( 0, iTab-1 );
			if( m_tabs.GetCount() > 0 )
				m_pOwner->OnTabSwitched( m_tabs[m_ActiveTab].dwID );
			else
            {
                // ..and the last one
                m_pOwner->OnTabSwitched( 0 );
                m_ActiveTab = -1;
            }
		}
		else if ( iTab < m_ActiveTab )
			m_ActiveTab--;
	}

    m_LastTab = std::min( m_LastTab, int( m_tabs.GetCount() - 1 ) );
	m_bPositionsDirty = true;
	InvalidateRect( NULL );
}

void CVSTabs::SetTabUserData ( DWORD dwID, void* pUserData )
{
    TabInfo* pTab = FindTabByID( dwID );
    if( !pTab )
        return;
    pTab->pUserData = pUserData;
}

void* CVSTabs::GetTabUserData ( DWORD dwID )
{
    TabInfo* pTab = FindTabByID( dwID );
    if( !pTab )
        return NULL;
    return pTab->pUserData;
}

DWORD CVSTabs::GetActiveTabID()
{
    if( m_ActiveTab < 0 )
        return 0;
    return m_tabs[m_ActiveTab].dwID;
}

void CVSTabs::CycleTab( bool bCycleForward )
{
    if( m_tabs.GetCount() < 2 )
        return;

    int iNewTabIndex = m_ActiveTab;
    if( bCycleForward )
    {
        iNewTabIndex++;
        if( iNewTabIndex >= int( m_tabs.GetCount() ) )
            iNewTabIndex = 0;
    }
    else
    {
        iNewTabIndex--;
        if( iNewTabIndex < 0 )
            iNewTabIndex = int( m_tabs.GetCount() ) - 1;
    }
    SwitchToTab( m_tabs[iNewTabIndex].dwID );
}

HRESULT CVSTabs::GetAllTabs( CAtlArray<DWORD> &arrTabIDs )
{
    if( !arrTabIDs.SetCount( m_tabs.GetCount() ) )
        return E_OUTOFMEMORY;
    for( size_t i=0; i<m_tabs.GetCount(); i++ )
        arrTabIDs[i] = m_tabs[i].dwID;
    return S_OK;
}