#pragma once
#include "MouseProcessor.h"
#include "VSTabsIface.h"

// VisualStudio-like tabs panel, working effectively enough to be used inside the remote desktop.
// The two fonts for the text labels are the system font for menus + "Normal" weight,
//      and the system font for menus + "semi-bold" weight
// The height of the control is the system menu height + 1 pixel.
// The right-to-left layout is untested.
// The individual tabs are referred by DWORD TabID, which is actually a tab serial number.
// The 1-st tab created has ID '1', the 0 is considered invalid tab ID.
class CVSTabs:
	public CWindowImpl<CVSTabs, CWindow, CControlWinTraits>,
	public CMouseProcessor
{
public:
	CVSTabs(void);
	~CVSTabs(void);

	// ==== Interface functions ====
	HRESULT Create( CWindow* pParent, IVSTabsHost* pOwner, UINT uID );

	// Add a new tab, return the newly added tab ID.
    DWORD AddTab( LPCTSTR strText, void* pUserData = NULL );

    // Get the tab ID of the active tab, or 0 if there're no tabs.
    DWORD GetActiveTabID();

    void SetTabUserData( DWORD dwID, void* pUserData );

    // Get the tab user data set by SetTabUserData or passed to AddTab
    // Returns NULL if the passed ID is invalid.
    void* GetTabUserData( DWORD dwID );

	void SwitchToTab( DWORD dwID );

	void RemoveTab( DWORD dwID );

    void CycleTab( bool bCycleForward );

	void SetTabTooltip( DWORD dwID, LPCTSTR strTooltipText );

	void SetTabText( DWORD dwID, LPCTSTR strText );

    HRESULT GetAllTabs( CAtlArray<DWORD> &arrTabIDs );

	inline void Resize()
	{
		CRect rc;
		GetParent().GetClientRect( &rc );
		rc.bottom = rc.top + m_TotalHeight;
		Resize( rc );
	}
	inline void Resize( CRect& rc )
	{
		MoveWindow( &rc, FALSE );
	}


	DECLARE_WND_CLASS_EX( L"VsTabs", 0, COLOR_BTNFACE )
	BEGIN_MSG_MAP( CVSTabs )
		MESSAGE_HANDLER(WM_CREATE, OnCreate)
		MESSAGE_HANDLER(WM_SIZE, OnSize)
		MESSAGE_HANDLER(WM_DESTROY, OnDestroy)
        MESSAGE_HANDLER(WM_PAINT, OnPaint)

		CHAIN_MSG_MAP(CMouseProcessor)
	END_MSG_MAP()
protected:

	// -1       nothing interesting in that point
	// -2       'close' button
	// -3       'right' button if visible,
	// -4       'left' button if visible
	// >=0      zero-based index of tab
	virtual int HitTest(CPoint pt);

	// Messages from the mouse for individual controls.
	// iCtrl meaning is same as in case of HitTest
	virtual void OnMouseInOut( int iCtrl, bool bIn );
	virtual void OnMouseHover( int iCtrl, CPoint pt );
	virtual void OnMouseDown( int iCtrl, DWORD vk );
	virtual void OnMouseUp( int iCtrl, DWORD vk );
	virtual void OnMouseClick( int iCtrl, DWORD vk );
	
	// This class holds an information about a single tab.
    // It also provides methods for painting and hit testing.
    // All the coordinates are from the upper-left corner of the window client area.
	// An instance of this class usually owns HRGN GDI resource.
    class __declspec(novtable) TabInfo
	{
	public:
		TabInfo();
		~TabInfo();
		inline TabInfo(const TabInfo& that)
		{
			CopyFrom(that);
		}
		inline TabInfo& operator=(const TabInfo& that)
		{
			CopyFrom(that);
			return *this;
		}

		// Paint a region with currently selected brush.
        // Only used for the active tab.
		bool FillTabRegion( HDC dc, HBRUSH br ) const;

		// Draw a frame around the tab with currently selected pen, and print the caption text.
		// SetTextAlign should be set to TA_CENTER | TA_BOTTOM
		bool DrawFrameAndText(HDC dc, bool bLeftSideOverlapped) const;

		// Recalculate the border shape and a region
		// Return value is the leftmost point of the next tab
		// cy=height of the control
		// xLeftMost (in/out) - x-coordinate of the leftmost point of this tab (in) and of the next tab (out)
		// xEnd (out) - x-coordinate of the end of the tab
		bool RecalcLayout(int cy, int *xLeftMost, int* xEnd);

		bool RecalcRegion() const;

		// The ACTIVE font mush be selected in the DC
		bool SetText( HDC dc, LPCTSTR strNewText );
		inline CString GetText() const { return Caption; }

		int GetWidth( bool bIncludeOverflow ) const;

		int GetRight() const;

		// Returns a bounding box,
        // possibly including the area overflowed by the next tab where border[0].x <= x < border[1].x
		CRect GetRect( bool bIncludeOverflow ) const;

		// Returns true if the point lies within the region
        bool HitTest( CPoint client ) const;

		// Tab ID, that is exposed outside.
        DWORD dwID;
        void* pUserData;
	protected:
		inline void CopyFrom( const TabInfo& that );

		CString Caption;
		int pxCaptionLength;

		CString Tooltip;

		// 7 points @ the edge:
		//          [3]-----------------------[4]
		//      [2]                             [5]
		//     [1]                               |
		//    /                                  |
		//   /                                   |
		// [0]----------------------------------[6]
		CPoint Border[7];

		// Region: used to paint an active tab with white color, and for hit tests.
		mutable HRGN Region;

		// Get the x-coordinate of the center of the text
		int GetTextCenter() const;
	};

	// Various GDI handles for painting:
	HFONT m_fontInactive, m_fontActive; // < Fonts for active and inactive caption
	HBRUSH m_brActiveTab;   //< Inactive tabs are not filled at all
	HPEN m_pnOutline;       //< Pen to outline tabs
	DWORD m_colText;        //< Text color

	// Height of the control (scaled with the system menu height in display properties)
	int m_TotalHeight;

	// Size of the buttons
	CSize m_szButtons;

	// Some data for 3 buttons.
    // Often the close button is the only visible.
    // They all are painted using the DrawFrameControl() API
	struct ButtonInfo
	{
		inline ButtonInfo(UINT dfcType, UINT dfcFlag) : m_dfcType(dfcType), m_dfcFlag(dfcFlag)
		{
			m_bVisible = m_bEnabled = m_bOver = m_bDown = false;
		}

		void RecalcLayout( CSize sz, int iRight, CPoint ptOffset );

        // Call DrawFrameControl
        // The type and state are derived from the class data members
        void Draw( HDC dc ) const;

		CRect m_rc;
		bool m_bVisible;
		bool m_bEnabled;
		bool m_bOver;
		bool m_bDown;

		const UINT m_dfcType, m_dfcFlag;
	};
	ButtonInfo m_btnLeft, m_btnRight, m_btnClose;

	static const int s_Buttons = 3;
	ButtonInfo *m_pButtons[s_Buttons];

	// true means that the geometry of tabs needs recalculating during the next WM_PAINT operation
	bool m_bPositionsDirty;

	// Index of the leftmost and rightmost visible tab
	int m_FirstTab, m_LastTab;
    // Index of the active tab
    int m_ActiveTab;
	// Is set when even the single active tab doesn't fit the width
	bool m_ActiveTabNeedsShrinkage;

	// The actual array storing the tabs
    CAtlArray<TabInfo> m_tabs;

    // The hash map for the fast lookups
    // The key is the ID, the value is the index in m_tabs array;
    // so when removing a tab, the indices needs to be recalculated.
    typedef CAtlMap<DWORD, int> MapType;
    MapType m_tabsLookup;

	IVSTabsHost* m_pOwner;

	DWORD m_dwNextTabID;

	LRESULT OnCreate(UINT nMsg, WPARAM wParam,  LPARAM lParam, BOOL& bHandled);
	LRESULT OnSize(UINT nMsg, WPARAM wParam,  LPARAM lParam, BOOL& bHandled);
	LRESULT OnDestroy(UINT nMsg, WPARAM wParam,  LPARAM lParam, BOOL& bHandled);
	LRESULT OnPaint(UINT nMsg, WPARAM wParam,  LPARAM lParam, BOOL& bHandled);

	inline CRect IntersectRect(CRect& r1, CRect& r2)
	{
		CRect r;
		r.IntersectRect(r1, r2);
		return r;
	}
private:
	// [re]calculate text 
	BOOL RecalcMetrics();
	void RecalcLayout();

	TabInfo* FindTabByID( DWORD dwID )
	{
        const MapType::CPair* p = m_tabsLookup.Lookup( dwID );
        if(!p) return NULL;
        return &m_tabs[ p->m_value ];
	}
};
